import abc
from dataclasses import dataclass
from enum import Enum
from typing import Generic, List, Optional, Type, Union

from pure_protobuf.dataclasses_ import field, message, one_of, part
from pure_protobuf.oneof import OneOf_

from ... import base, objects
from .common import ClassT, ProtobufSerializer, types_registry, types_registry_reversed, deserialize_one_of_field


@message
@dataclass
class LocalizedStringItem:
    lang: str = field(1)
    text: str = field(2)


@message
@dataclass
class LocalizedString(ProtobufSerializer[base.LocalizedString]):
    items: List[LocalizedStringItem] = field(1)

    @staticmethod
    def serialize(obj: base.LocalizedString) -> 'LocalizedString':
        return LocalizedString([LocalizedStringItem(lang, text) for lang, text in obj.lang_to_text.items()])

    def deserialize(self) -> base.LocalizedString:
        return base.LocalizedString(lang_to_text={entry.lang: entry.text for entry in self.items})


# For library string enumerations
@message
@dataclass
class StrEnum(ProtobufSerializer[Enum], Generic[ClassT]):
    value: str = field(1)

    @staticmethod
    @abc.abstractmethod
    def enum_cls() -> Type[Enum]:
        ...

    @classmethod
    def serialize(cls, obj: Enum) -> 'StrEnum':
        return cls(value=obj.value)

    def deserialize(self) -> Enum:
        return self.enum_cls()(self.value)


@message
class TextFormat(StrEnum[base.TextFormat]):
    @staticmethod
    def enum_cls() -> Type[base.TextFormat]:
        return base.TextFormat


@message
@dataclass
class Title(ProtobufSerializer[base.Title]):
    text: LocalizedString = field(1)
    format: TextFormat = field(2)

    @staticmethod
    def serialize(obj: base.Title) -> 'Title':
        return Title(text=LocalizedString.serialize(obj.text), format=TextFormat.serialize(obj.format))

    def deserialize(self) -> base.Title:
        return base.Title(text=self.text.deserialize(), format=self.format.deserialize())


@message
class SbSChoice(StrEnum[base.SbSChoice]):
    @staticmethod
    def enum_cls() -> Type[base.SbSChoice]:
        return base.SbSChoice


@message
@dataclass
class BinaryEvaluation(ProtobufSerializer[base.BinaryEvaluation]):
    ok: bool = field(1)

    @staticmethod
    def serialize(obj: base.BinaryEvaluation) -> 'BinaryEvaluation':
        return BinaryEvaluation(obj.ok)

    def deserialize(self) -> base.BinaryEvaluation:
        return base.BinaryEvaluation(self.ok)


@message
@dataclass
class Metadata(ProtobufSerializer[base.Metadata]):
    metadata: str = field(1)

    @staticmethod
    def serialize(obj: base.Metadata) -> 'Metadata':
        return Metadata(obj.metadata)

    def deserialize(self) -> base.Metadata:
        return base.Metadata(self.metadata)


@message
@dataclass
class Text(ProtobufSerializer[objects.Text]):
    text: str = field(1)

    @staticmethod
    def serialize(obj: objects.Text) -> 'Text':
        return Text(obj.text)

    def deserialize(self) -> objects.Text:
        return objects.Text(self.text)


@message
@dataclass
class Audio(ProtobufSerializer[objects.Audio]):
    url: str = field(1)

    @staticmethod
    def serialize(obj: objects.Audio) -> 'Audio':
        return Audio(obj.url)

    def deserialize(self) -> objects.Audio:
        return objects.Audio(self.url)


@message
@dataclass
class Image(ProtobufSerializer[objects.Image]):
    url: str = field(1)

    @staticmethod
    def serialize(obj: objects.Image) -> 'Image':
        return Image(obj.url)

    def deserialize(self) -> objects.Image:
        return objects.Image(self.url)


@message
@dataclass
class Video(ProtobufSerializer[objects.Video]):
    url: str = field(1)

    @staticmethod
    def serialize(obj: objects.Video) -> 'Video':
        return Video(obj.url)

    def deserialize(self) -> objects.Video:
        return objects.Video(self.url)


# For user-defined string enumerations, including Class inheritors
@message
@dataclass
class Class(ProtobufSerializer[base.Class]):
    value: str = field(1)
    cls: str = field(2)

    @staticmethod
    def serialize(obj: base.Class) -> 'Class':
        cls = types_registry[type(obj)]
        return Class(value=obj.value, cls=cls)

    def deserialize(self) -> base.Class:
        cls = types_registry_reversed[self.cls]
        return cls(self.value)


@message
@dataclass
class Label(ProtobufSerializer[base.Label]):
    label: OneOf_ = one_of(
        cls=part(Class, 1),
        sbs_choice=part(SbSChoice, 2),
        binary_evaluation=part(BinaryEvaluation, 3),
    )

    @staticmethod
    def serialize(obj: base.Label) -> 'Label':
        label = Label()
        if isinstance(obj, base.BinaryEvaluation):
            label.label.binary_evaluation = BinaryEvaluation.serialize(obj)
        elif isinstance(obj, base.SbSChoice):
            label.label.sbs_choice = SbSChoice.serialize(obj)
        elif isinstance(obj, base.Class):
            label.label.cls = Class.serialize(obj)
        else:
            assert False, f'unexpected Label type: {type(obj)}'
        return label

    def deserialize(self) -> base.Label:
        return deserialize_one_of_field(self.label)


@message
@dataclass
class Object(ProtobufSerializer[base.Object]):
    obj: OneOf_ = one_of(
        cls=part(Class, 1),
        sbs_choice=part(SbSChoice, 2),
        binary_evaluation=part(BinaryEvaluation, 3),
        metadata=part(Metadata, 4),
        text=part(Text, 5),
        audio=part(Audio, 6),
        image=part(Image, 7),
        video=part(Video, 8),
    )

    @staticmethod
    def serialize(obj: base.Object) -> 'Object':
        obj_ = Object()
        if isinstance(obj, base.BinaryEvaluation):
            obj_.obj.binary_evaluation = BinaryEvaluation.serialize(obj)
        elif isinstance(obj, base.SbSChoice):
            obj_.obj.sbs_choice = SbSChoice.serialize(obj)
        elif isinstance(obj, base.Class):
            obj_.obj.cls = Class.serialize(obj)
        elif isinstance(obj, base.Metadata):
            obj_.obj.metadata = Metadata.serialize(obj)
        elif isinstance(obj, objects.Text):
            obj_.obj.text = Text.serialize(obj)
        elif isinstance(obj, objects.Audio):
            obj_.obj.audio = Audio.serialize(obj)
        elif isinstance(obj, objects.Image):
            obj_.obj.image = Image.serialize(obj)
        elif isinstance(obj, objects.Video):
            obj_.obj.video = Video.serialize(obj)
        else:
            assert False, f'unexpected Object type: {type(obj)}'
        return obj_

    def deserialize(self) -> base.Object:
        return deserialize_one_of_field(self.obj)


# We could leave Object's one_of field unfilled, it would also make it None.
# But, we can't make things like List[Optional[Object]] in other serializer classes,
# so we have to create wrapper for "object or None".
@message
@dataclass
class OptionalObject(ProtobufSerializer[Optional[base.Object]]):
    obj: Optional[Object] = field(1, default=None)

    @staticmethod
    def serialize(obj: Optional[base.Object]) -> 'OptionalObject':
        return OptionalObject(obj=Object.serialize(obj) if obj else None)

    def deserialize(self) -> Optional[base.Object]:
        return self.obj.deserialize() if self.obj else None


@message
@dataclass
class ObjectMeta(ProtobufSerializer[base.ObjectMeta]):
    type: str = field(1)
    name: Optional[str] = field(2, default=None)
    title: Optional[Title] = field(3, default=None)
    required: bool = field(4, default=True)

    @staticmethod
    def serialize(obj: base.ObjectMeta) -> 'ObjectMeta':
        return ObjectMeta(
            type=types_registry[obj.type],
            name=obj.name,
            title=Title.serialize(obj.title) if obj.title else None,
            required=obj.required,
        )

    def deserialize(self) -> base.ObjectMeta:
        return base.ObjectMeta(
            type=types_registry_reversed[self.type],
            name=self.name,
            title=self.title.deserialize() if self.title else None,
            required=self.required,
        )


@message
@dataclass
class AvailableLabels(ProtobufSerializer[base.AvailableLabels]):
    labels: List[Label] = field(1)

    @staticmethod
    def serialize(obj: base.AvailableLabels) -> 'AvailableLabels':
        return AvailableLabels([Label.serialize(label) for label in obj.labels])

    def deserialize(self) -> base.AvailableLabels:
        return base.AvailableLabels([label.deserialize() for label in self.labels])


@message
@dataclass
class Consequence(ProtobufSerializer[base.Consequence]):
    consequence: OneOf_ = one_of(
        available_labels=part(AvailableLabels, 1),
    )

    @staticmethod
    def serialize(obj: base.Consequence) -> 'Consequence':
        consequence = Consequence()
        if isinstance(obj, base.AvailableLabels):
            consequence.consequence.available_labels = AvailableLabels.serialize(obj)
        else:
            raise ValueError(f'unexpected consequence type: {type(obj)}')
        return consequence

    def deserialize(self) -> base.Consequence:
        return deserialize_one_of_field(self.consequence)


@message
@dataclass
class ConditionEquals(ProtobufSerializer[base.ConditionEquals]):
    what: str = field(1)
    to: OneOf_ = one_of(
        str=part(str, 2),
    )

    @staticmethod
    def serialize(obj: base.ConditionEquals) -> 'ConditionEquals':
        condition = ConditionEquals(what=obj.what)
        if isinstance(obj.to, str):
            condition.to.str = obj.to
        else:
            raise ValueError(f'unsupported equals to type: {type(obj.to)}')
        return condition

    def deserialize(self) -> base.ConditionEquals:
        return base.ConditionEquals(what=self.what, to=deserialize_one_of_field(self.to))


@message
@dataclass
class Condition(ProtobufSerializer[base.Condition]):
    condition: OneOf_ = one_of(
        equals=part(ConditionEquals, 1),
    )

    @staticmethod
    def serialize(obj: base.Condition) -> 'Condition':
        condition = Condition()
        if isinstance(obj, base.ConditionEquals):
            condition.condition.equals = ConditionEquals.serialize(obj)
        else:
            raise ValueError(f'unexpected condition type: {type(obj)}')
        return condition

    def deserialize(self) -> base.Condition:
        return deserialize_one_of_field(self.condition)


# then/else branches are originally unions, better option to pack them to one_of, but it's not possible because union
# contains If itself and we can't pass class to one_of part using quotation syntax like 'If'
@dataclass
class If(ProtobufSerializer[base.If]):
    condition: Condition = field(1)
    then_if: Optional['If'] = field(2, default=None)
    then_consequence: Optional[Consequence] = field(3, default=None)
    else_if: Optional['If'] = field(4, default=None)
    else_consequence: Optional[Consequence] = field(5, default=None)

    def __post_init__(self):
        assert (self.then_if is not None) ^ (self.then_consequence is not None)
        assert not (self.else_if is not None and self.else_consequence is not None)

    @staticmethod
    def serialize(obj: base.If) -> 'If':
        return If(
            condition=Condition.serialize(obj.condition),
            then_if=If.serialize(obj.then) if obj.then and isinstance(obj.then, base.If) else None,
            then_consequence=Consequence.serialize(obj.then)
            if obj.then and isinstance(obj.then, base.Consequence)
            else None,
            else_if=If.serialize(obj.else_) if obj.else_ and isinstance(obj.else_, base.If) else None,
            else_consequence=Consequence.serialize(obj.else_)
            if obj.else_ and isinstance(obj.else_, base.Consequence)
            else None,
        )

    def deserialize(self) -> base.If:
        then = self.then_if.deserialize() if self.then_if else self.then_consequence.deserialize()
        else_ = None
        if self.else_if:
            else_ = self.else_if.deserialize()
        if self.else_consequence:
            else_ = self.else_consequence.deserialize()
        return base.If(condition=self.condition.deserialize(), then=then, else_=else_)


# self-referencing classes are decorated with @message separately, see https://github.com/eigenein/protobuf/issues/96
If = message(If)


@message
class LabelsDisplayType(StrEnum[base.LabelsDisplayType]):
    @staticmethod
    def enum_cls() -> Type[base.LabelsDisplayType]:
        return base.LabelsDisplayType


@message
@dataclass
class ClassMeta(ObjectMeta, ProtobufSerializer[base.ClassMeta]):
    available_labels: Optional[If] = field(100, default=None)
    input_display_type: LabelsDisplayType = field(101, default=LabelsDisplayType(base.LabelsDisplayType.MULTI))
    text_format: TextFormat = field(102, default=TextFormat(base.TextFormat.PLAIN))

    @staticmethod
    def serialize(obj: base.ClassMeta) -> 'ClassMeta':
        obj_meta = ObjectMeta.serialize(obj)
        return ClassMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            available_labels=If.serialize(obj.available_labels) if obj.available_labels else None,
            input_display_type=LabelsDisplayType.serialize(obj.input_display_type),
            text_format=TextFormat.serialize(obj.text_format),
        )

    def deserialize(self) -> base.ClassMeta:
        obj_meta = super(ClassMeta, self).deserialize()
        return base.ClassMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            available_labels=self.available_labels.deserialize() if self.available_labels else None,
            input_display_type=self.input_display_type.deserialize(),
            text_format=self.text_format.deserialize(),
        )


@message
@dataclass
class TextValidation(ProtobufSerializer[objects.TextValidation]):
    regex: LocalizedString = field(1)
    hint: LocalizedString = field(2)

    @staticmethod
    def serialize(obj: objects.TextValidation) -> 'TextValidation':
        return TextValidation(
            regex=LocalizedString.serialize(obj.regex),
            hint=LocalizedString.serialize(obj.hint),
        )

    def deserialize(self) -> objects.TextValidation:
        return objects.TextValidation(
            regex=self.regex.deserialize(),
            hint=self.hint.deserialize(),
        )


@message
@dataclass
class TextMeta(ObjectMeta, ProtobufSerializer[objects.TextMeta]):
    format: TextFormat = field(100, default=TextFormat(base.TextFormat.PLAIN))
    validation: Optional[TextValidation] = field(101, default=None)

    @staticmethod
    def serialize(obj: objects.TextMeta) -> 'TextMeta':
        obj_meta = ObjectMeta.serialize(obj)
        return TextMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            format=TextFormat.serialize(obj.format),
            validation=TextValidation.serialize(obj.validation) if obj.validation else None,
        )

    def deserialize(self) -> objects.TextMeta:
        obj_meta = super(TextMeta, self).deserialize()
        return objects.TextMeta(
            type=obj_meta.type,
            name=obj_meta.name,
            title=obj_meta.title,
            required=obj_meta.required,
            format=self.format.deserialize(),
            validation=self.validation.deserialize() if self.validation else None,
        )


@message
@dataclass
class ObjectMetaT(ProtobufSerializer[base.ObjectMeta]):
    meta: OneOf_ = one_of(
        object=part(ObjectMeta, 1),
        cls=part(ClassMeta, 2),
        text=part(TextMeta, 3),
    )

    @staticmethod
    def serialize(obj: base.ObjectMeta) -> 'ObjectMetaT':
        meta = ObjectMetaT()
        if isinstance(obj, base.ClassMeta):
            meta.meta.cls = ClassMeta.serialize(obj)
        elif isinstance(obj, objects.TextMeta):
            meta.meta.text = TextMeta.serialize(obj)
        else:
            meta.meta.object = ObjectMeta.serialize(obj)
        return meta

    def deserialize(self) -> base.ObjectMeta:
        return deserialize_one_of_field(self.meta)


@message
@dataclass
class FunctionArgument(ProtobufSerializer[Union[Type[base.Object], base.ObjectMeta]]):
    arg: OneOf_ = one_of(
        type=part(str, 1),
        meta=part(ObjectMetaT, 2),
    )

    @staticmethod
    def serialize(obj: Union[Type[base.Object], base.ObjectMeta]) -> 'FunctionArgument':
        arg = FunctionArgument()
        if isinstance(obj, base.ObjectMeta):
            arg.arg.meta = ObjectMetaT.serialize(obj)
        else:
            arg.arg.type = types_registry[obj]
        return arg

    def deserialize(self) -> Union[Type[base.Object], base.ObjectMeta]:
        which_one = self.arg.which_one_of
        if which_one == 'type':
            return types_registry_reversed[self.arg.type]
        else:
            return self.arg.meta.deserialize()


@message
@dataclass
class ClassificationFunction(ProtobufSerializer[base.ClassificationFunction]):
    inputs: List[FunctionArgument] = field(1)
    cls: FunctionArgument = field(2)

    @staticmethod
    def serialize(obj: base.ClassificationFunction) -> 'ClassificationFunction':
        return ClassificationFunction(
            inputs=[FunctionArgument.serialize(input) for input in obj.inputs],
            cls=FunctionArgument.serialize(obj.cls),
        )

    def deserialize(self) -> base.ClassificationFunction:
        return base.ClassificationFunction(
            inputs=tuple(FunctionArgument.deserialize(input) for input in self.inputs),
            cls=FunctionArgument.deserialize(self.cls),
        )


sbs_choice = FunctionArgument()
sbs_choice.arg.type = types_registry[base.SbSChoice]


@message
@dataclass
class SbSFunction(ProtobufSerializer[base.SbSFunction]):
    inputs: List[FunctionArgument] = field(1)
    hints: List[FunctionArgument] = field(2, default_factory=list)
    choice: FunctionArgument = field(3, default=sbs_choice)

    @staticmethod
    def serialize(obj: base.SbSFunction) -> 'SbSFunction':
        return SbSFunction(
            inputs=[FunctionArgument.serialize(input) for input in obj.inputs],
            hints=[FunctionArgument.serialize(hint) for hint in obj.hints] if obj.hints else [],
            choice=FunctionArgument.serialize(obj.choice),
        )

    def deserialize(self) -> base.SbSFunction:
        return base.SbSFunction(
            inputs=tuple(FunctionArgument.deserialize(input) for input in self.inputs),
            hints=tuple(FunctionArgument.deserialize(hint) for hint in self.hints) if self.hints else None,
            choice=FunctionArgument.deserialize(self.choice),
        )


@message
@dataclass
class AnnotationFunction(ProtobufSerializer[base.AnnotationFunction]):
    inputs: List[FunctionArgument] = field(1)
    outputs: List[FunctionArgument] = field(2)
    evaluation: FunctionArgument = field(3)

    @staticmethod
    def serialize(obj: base.AnnotationFunction) -> 'AnnotationFunction':
        return AnnotationFunction(
            inputs=[FunctionArgument.serialize(input) for input in obj.inputs],
            outputs=[FunctionArgument.serialize(output) for output in obj.outputs],
            evaluation=FunctionArgument.serialize(obj.evaluation),
        )

    def deserialize(self) -> base.AnnotationFunction:
        return base.AnnotationFunction(
            inputs=tuple(FunctionArgument.deserialize(input) for input in self.inputs),
            outputs=tuple(FunctionArgument.deserialize(output) for output in self.outputs),
            evaluation=FunctionArgument.deserialize(self.evaluation),
        )


@message
@dataclass
class TaskFunction(ProtobufSerializer[base.TaskFunction]):
    function: OneOf_ = one_of(
        classification=part(ClassificationFunction, 1),
        sbs=part(SbSFunction, 2),
        annotation=part(AnnotationFunction, 3),
    )

    @staticmethod
    def serialize(obj: base.TaskFunction) -> 'TaskFunction':
        function = TaskFunction()
        if isinstance(obj, base.ClassificationFunction):
            function.function.classification = ClassificationFunction.serialize(obj)
        elif isinstance(obj, base.SbSFunction):
            function.function.sbs = SbSFunction.serialize(obj)
        elif isinstance(obj, base.AnnotationFunction):
            function.function.annotation = AnnotationFunction.serialize(obj)
        else:
            raise ValueError(f'unexpected function {obj}')
        return function

    def deserialize(self) -> base.TaskFunction:
        return deserialize_one_of_field(self.function)


@message
@dataclass
class TaskSpec(ProtobufSerializer[base.TaskSpec]):
    id: str = field(1)
    function: TaskFunction = field(2)
    name: LocalizedString = field(3)
    description: LocalizedString = field(4)
    instruction: LocalizedString = field(5)

    @staticmethod
    def serialize(obj: base.TaskSpec) -> 'TaskSpec':
        return TaskSpec(
            id=obj.id,
            function=TaskFunction.serialize(obj.function),
            name=LocalizedString.serialize(obj.name),
            description=LocalizedString.serialize(obj.description),
            instruction=LocalizedString.serialize(obj.instruction),
        )

    def deserialize(self) -> base.TaskSpec:
        return base.TaskSpec(
            id=self.id,
            function=self.function.deserialize(),
            name=self.name.deserialize(),
            description=self.description.deserialize(),
            instruction=self.instruction.deserialize(),
        )



import difflib
import logging
from typing import Tuple, Optional

from bs4 import BeautifulSoup
import toloka.client as toloka

from .. import base, experts, mapping, instruction, project, task_spec as spec

from .utils import ask

logger = logging.getLogger(__name__)


# Toloka localized projects are not suitable for us, because:
# - in task view, only labels can be localized in normal way without if/else conditions, i.e. we can't change output
#   validation depends on language for example, see
#   https://toloka.ai/ru/docs/guide/concepts/project-languages.html#project-languages__interface-translate.
# - Worker can choose multiple languages, on which he can perform tasks, but Toloka chooses instruction language
#   depends on worker language in settings, not language in pool filters. For tasks with language-specific
#   instructions it can cause problems.
def get_task_id(task_spec: spec.PreparedTaskSpec):
    if task_spec.scenario == project.Scenario.DEFAULT:
        id = task_spec.id
    else:
        id = f'{task_spec.id}_{task_spec.scenario.value}'
    return f'{id}_{task_spec.lang}'


def _log_found_project(project_id: str, task_spec: spec.PreparedTaskSpec):
    logger.debug(f'suitable Toloka project {project_id} is found for {task_spec.log_message()}')


def define_task(task_spec: spec.PreparedTaskSpec, client: toloka.TolokaClient, interactive: bool = True):
    if task_spec.scenario == project.Scenario.DEFAULT and isinstance(task_spec, spec.AnnotationTaskSpec):
        define_task(task_spec.check, client, interactive)

    found_project = find_project(task_spec, client)
    if found_project is None:
        prj = toloka.Project(private_comment=get_task_id(task_spec))
    else:
        prj, diff = found_project
        if not diff:
            _log_found_project(prj.id, task_spec)
            logger.info('no changes in task')
            return
        if interactive and not ask(f'{diff}update Toloka project {prj.id} according to the changes shown above?'):
            logger.info('task update is canceled')
            return

    task_spec.task_mapping.validate()

    prj.public_name = task_spec.name[task_spec.lang]
    prj.public_description = task_spec.description[task_spec.lang]
    prj.public_instructions = get_instruction(task_spec)
    prj.task_spec = toloka.project.task_spec.TaskSpec(
        input_spec=task_spec.task_mapping.to_toloka_input_spec(),
        output_spec=task_spec.task_mapping.to_toloka_output_spec(),
        view_spec=task_spec.view,
    )

    if found_project is None:
        logger.debug(f'creating Toloka project for {task_spec.log_message()}')
        client.create_project(prj)
        logger.info('task is defined')
    else:
        logger.debug(f'updating Toloka project {prj.id} for {task_spec.log_message()}')
        client.update_project(prj.id, prj)
        logger.info('task is updated')
        if interactive:
            notify_workers_about_task_change(task_spec, prj, client)


def find_project(
    task_spec: spec.PreparedTaskSpec,
    client: toloka.TolokaClient,
) -> Optional[Tuple[toloka.Project, str]]:
    exists_msg = f'Toloka project is exists for {task_spec.log_message()}'

    for prj in client.get_projects(status='ACTIVE'):
        if not prj.private_comment == get_task_id(task_spec):
            continue
        if not mapping.check_project_is_suitable(prj, task_spec.task_mapping):
            raise ValueError(
                f'{exists_msg}, but function of task is changed in incompatible way. Use different task ID'
            )
        diff = get_project_diff(prj, task_spec)
        diff_msg = f'{exists_msg} and differs from current task spec:\n\n{diff}' if diff else ''
        return prj, diff_msg
    return None


def get_project_diff(prj: toloka.Project, task_spec: spec.PreparedTaskSpec) -> str:
    assert prj.private_comment == get_task_id(task_spec)
    diffs = []
    for current_value, existing_value, value_display, value_name in (
        (task_spec.name[task_spec.lang], prj.public_name, None, 'name'),
        (task_spec.description[task_spec.lang], prj.public_description, None, 'description'),
        (get_instruction(task_spec), prettify_html(prj.public_instructions), None, 'instruction'),
        (
            task_spec.view.config.to_json(pretty=True),
            prj.task_spec.view_spec.config.to_json(pretty=True),
            lambda config: config.to_json(pretty=True),
            'view',
        ),
    ):
        if current_value != existing_value:
            diffs.append(f'{value_name} changed:\n\n{string_diff(existing_value, current_value)}\n\n')
    return '\n'.join(diffs)


def check_project(task_spec: spec.PreparedTaskSpec, client: toloka.TolokaClient) -> toloka.Project:
    found_project = find_project(task_spec, client)
    if found_project is None:
        raise RuntimeError(f'task "{task_spec.id}" is not defined')
    prj, diff = found_project
    if diff:
        raise RuntimeError(diff)
    _log_found_project(prj.id, task_spec)
    return prj


# Toloka supports not all HTML tags and also changes HTML while creating/updating project, using escaping and some
# strange formatting. So we try so compare prettified versions of current and Toloka project instructions. It may break
# if Toloka begins to preprocess HTML in way that our prettifier could not format instructions in the same way. In such
# situation, we will have to put current instruction or its hash to some field in Toloka project, i.e. to private
# comment or metadata, and do not compare with instruction directly.
def prettify_html(html: str) -> str:
    return BeautifulSoup(html, 'html.parser').prettify()


def string_diff(s1: str, s2: str) -> str:
    diff = difflib.unified_diff(s1.splitlines(), s2.splitlines(), lineterm='')
    return '\n'.join(diff_line.rstrip() for diff_line in diff)


def get_instruction(task_spec: spec.PreparedTaskSpec) -> str:
    lang = task_spec.lang
    worker = '<div>' + task_spec.instruction[lang] + '</div>'
    acceptance = instruction.acceptance[lang]

    if isinstance(task_spec.function, base.AnnotationFunction) and not isinstance(task_spec, spec.AnnotationTaskSpec):
        worker = '<div>' + instruction.annotation_check_cut[lang] + worker + '</div>'

    if task_spec.scenario == project.Scenario.DEFAULT:
        return prettify_html(instruction.css + worker + acceptance)

    worker_header, expert_cut = instruction.worker_header[lang], instruction.expert_cut[lang]

    has_evaluation = False
    if isinstance(task_spec.function, base.AnnotationFunction):
        has_evaluation = task_spec.scenario == project.Scenario.EXPERT_LABELING_OF_TASKS

    expert = instruction.get_formatted_expert_instruction(task_spec.function, lang, has_evaluation)

    return prettify_html(instruction.css + worker_header + worker + '<div>' + expert_cut + expert + '</div>')


def notify_workers_about_task_change(
    task_spec: spec.PreparedTaskSpec,
    prj: toloka.Project,
    client: toloka.TolokaClient,
):
    recipients_filter, recipients_msg = define_notification_recipients(task_spec, prj, client)
    if recipients_filter is None:
        return

    if not ask(f'notify workers of changes in the task? ({recipients_msg})'):
        return

    # If the name of the task has changed, the topic will contain a new name, which may confuse the workers.
    # In this case, we recommend writing in the message about the name change and specify the old name.
    topic = (
        base.LocalizedString(
            {
                'EN': 'Changes in task',
                'RU': 'Изменения в задании',
                'TR': 'Görevlerdeki değişiklikler',
                'KK': 'Тапсырмалардағы өзгерістер',
            }
        )
        + ' "'
        + task_spec.name
        + '"'
    )[task_spec.lang]

    message_lang = f'in {task_spec.lang}'
    if task_spec.lang != base.DEFAULT_LANG:
        message_lang += f' or {base.DEFAULT_LANG}'
    message = input(f'specify message for workers ({message_lang}): ')

    send_message_to_workers(topic, message, task_spec.lang, recipients_filter, client)

    logger.info('workers are notified')


def define_notification_recipients(
    task_spec: spec.PreparedTaskSpec,
    prj: toloka.Project,
    client: toloka.TolokaClient,
) -> Tuple[Optional[toloka.filter.FilterCondition], str]:
    if task_spec.scenario != project.Scenario.DEFAULT:
        skills = experts.get_skills(task_spec.id, task_spec.lang, client)
        if len(skills) == 0:
            logger.info('no registered experts to receive notification')
            return None, ''
        worker_filters = [
            toloka.filter.Skill(key=skill.id, operator=toloka.primitives.operators.CompareOperator.GTE, value=0.0)
            for _, skill in skills
        ]
        recipients_msg = (
            f'all experts who registered for task "{task_spec.id}" in {task_spec.lang} language '
            f'will receive a notification'
        )
    else:
        worker_filters = [toloka.filter.Languages.in_(task_spec.lang)]
        training_requirement = experts.find_training(prj.id, client)
        recipients_msg = (
            f'all {task_spec.lang} workers who have ever performed tasks associated with your Toloka account '
            f'will receive a notification'
        )
        if training_requirement:
            skill_id = experts.get_skill_id_from_training(client.get_pool(training_requirement.id))
            # user skill for training is deleted after retry_training_after_days
            worker_filters.append(
                toloka.filter.Skill(key=skill_id, operator=toloka.primitives.operators.CompareOperator.GTE, value=0.0)
            )
            recipients_msg = (
                f'all workers who have performed the task "{task_spec.id}" in {task_spec.lang} language within '
                f'the last {training_requirement.retry_training_after_days} days will receive a notification'
            )
    return toloka.filter.FilterAnd(worker_filters), recipients_msg


def send_message_to_workers(
    topic: str, message: str, lang: str, recipients_filter: toloka.filter.FilterCondition, client: toloka.TolokaClient
):
    try:
        logger.debug(f'sending message\n\ttopic: {topic}\n\ttext: {message}\n\trecipients filter: {recipients_filter}')
        client.compose_message_thread(
            topic={lang: topic},
            text={lang: message},
            recipients_select_type=toloka.message_thread.RecipientsSelectType.FILTER,
            recipients_filter=recipients_filter,
            answerable=False,
        )
    except toloka.exceptions.IncorrectActionsApiError as e:
        if e.code == 'INCORRECT_RECIPIENTS':
            logger.warn('no recipients for the notification')
        else:
            raise



from dataclasses import dataclass
from typing import List, Optional, Tuple

from lzy.api.v1 import op
from lzy.api.v1.whiteboard import whiteboard

from .. import classification, classification_loop, mapping, pool as pool_config
from .serialization import (
    TaskSpec,
    ObjectsList,
    TaskSingleSolutionList,
    Params,
    TolokaProject,
    TolokaPool,
    TolokaAssignment,
    Results,
    WorkerWeights,
)


# WARNING: field name changes are not backward-compatible
@dataclass
@whiteboard(tags=['v1.classification'], namespace='default')
class ClassificationWhiteboard:
    task_spec: TaskSpec
    lang: Optional[str]
    input_objects: ObjectsList
    control_objects: TaskSingleSolutionList
    params: Params

    project: TolokaProject

    pool_id: str = None

    pool: TolokaPool = None  # with final attrs after closing
    assignments: List[TolokaAssignment] = None

    results: Results = None
    worker_weights: Optional[WorkerWeights] = None


@op
def create_pool(
    loop: classification_loop.ClassificationLoop,
    control_objects: List[mapping.TaskSingleSolution],
    pool_cfg: pool_config.ClassificationConfig,
) -> str:
    return loop.create_pool(control_objects, pool_cfg).id


@op
def add_input_objects(
    loop: classification_loop.ClassificationLoop,
    input_objects: List[mapping.Objects],
    pool_id: str,
) -> None:
    # here and below we cast pool_id to str because lzy send value to @op in some wrapper which not JSON serializable
    # for toloka-kit requests
    loop.add_input_objects(str(pool_id), input_objects)


@op
def run_loop(
    loop: classification_loop.ClassificationLoop,
    pool_id: str,
) -> None:
    loop.loop(str(pool_id))


@op
def get_results(
    loop: classification_loop.ClassificationLoop,
    input_objects: List[mapping.Objects],
    pool_id: str,
) -> Tuple[classification.Results, Optional[classification.WorkerWeights]]:
    return loop.get_results(str(pool_id), input_objects)



from .base import *  # noqa
from .characteristics import *  # noqa
from .client import *  # noqa
from .params import *  # noqa
from .event_loop import *  # noqa
from .util import *  # noqa
from .ui import *  # noqa



from dataclasses import dataclass
from typing import List, Optional

from pure_protobuf.dataclasses_ import field, message, one_of, part
from pure_protobuf.oneof import OneOf_
import toloka.client as toloka

from ... import worker
from .common import ProtobufSerializer, T, deserialize_one_of_field, func_registry, load_func
from .toloka import TolokaSkill, TolokaFilterCondition


@message
@dataclass
class ExpertFilter(ProtobufSerializer[worker.ExpertFilter]):
    skills: List[TolokaSkill] = field(1)

    @staticmethod
    def serialize(obj: worker.ExpertFilter) -> 'ExpertFilter':
        return ExpertFilter([TolokaSkill.serialize(skill) for skill in obj.skills])

    def deserialize(self) -> worker.ExpertFilter:
        return worker.ExpertFilter([skill.deserialize() for skill in self.skills])


@message
@dataclass
class LanguageRequirement(ProtobufSerializer[worker.LanguageRequirement]):
    lang: str = field(1)
    verified: bool = field(2)

    @staticmethod
    def serialize(obj: worker.LanguageRequirement) -> 'LanguageRequirement':
        return LanguageRequirement(lang=obj.lang, verified=obj.verified)

    def deserialize(self) -> worker.LanguageRequirement:
        return worker.LanguageRequirement(lang=self.lang, verified=self.verified)


@message
@dataclass
class BaseWorkerFilter(ProtobufSerializer[worker.BaseWorkerFilter]):
    training_score: Optional[int] = field(1, default=None)

    @staticmethod
    def serialize(obj: worker.BaseWorkerFilter) -> 'BaseWorkerFilter':
        return BaseWorkerFilter(training_score=obj.training_score)

    def deserialize(self) -> worker.BaseWorkerFilter:
        return worker.BaseWorkerFilter(training_score=self.training_score)


@message
@dataclass
class WorkerFilterParams(ProtobufSerializer[worker.WorkerFilter.Params]):
    langs: List[LanguageRequirement] = field(1, default=list)
    regions: List[int] = field(2, default_factory=list)
    age_min: Optional[int] = field(3, default=None)
    age_max: Optional[int] = field(4, default=None)
    client_types: List[str] = field(5, default_factory=list)

    @staticmethod
    def serialize(obj: worker.WorkerFilter.Params) -> 'WorkerFilterParams':
        return WorkerFilterParams(
            langs=[LanguageRequirement.serialize(lang) for lang in sorted(obj.langs)],
            regions=list(obj.regions),
            age_min=obj.age_range[0],
            age_max=obj.age_range[1],
            client_types=[t.value for t in obj.client_types],
        )

    def deserialize(self) -> worker.WorkerFilter.Params:
        return worker.WorkerFilter.Params(
            langs={lang.deserialize() for lang in self.langs},
            regions=set(self.regions),
            age_range=(self.age_min, self.age_max),
            client_types={toloka.filter.ClientType.ClientType(t) for t in self.client_types},
        )


@message
@dataclass
class WorkerFilter(BaseWorkerFilter, ProtobufSerializer[worker.WorkerFilter]):
    filters: List[WorkerFilterParams] = field(100, default_factory=list)

    @staticmethod
    def serialize(obj: worker.WorkerFilter) -> 'WorkerFilter':
        base_filter = BaseWorkerFilter.serialize(obj)
        return WorkerFilter(
            training_score=base_filter.training_score, filters=[WorkerFilterParams.serialize(f) for f in obj.filters]
        )

    def deserialize(self) -> worker.WorkerFilter:
        base_filter = super(WorkerFilter, self).deserialize()
        return worker.WorkerFilter(
            training_score=base_filter.training_score,
            filters=[f.deserialize() for f in self.filters],
        )


@message
@dataclass
class CustomWorkerFilter(BaseWorkerFilter, ProtobufSerializer[worker.CustomWorkerFilter]):
    filter: Optional[TolokaFilterCondition] = field(100, default=None)

    @staticmethod
    def serialize(obj: worker.CustomWorkerFilter) -> 'CustomWorkerFilter':
        base_filter = BaseWorkerFilter.serialize(obj)
        return CustomWorkerFilter(
            training_score=base_filter.training_score,
            filter=TolokaFilterCondition.serialize(obj.filter) if obj.filter else None,
        )

    def deserialize(self) -> worker.CustomWorkerFilter:
        base_filter = super(CustomWorkerFilter, self).deserialize()
        return worker.CustomWorkerFilter(
            training_score=base_filter.training_score,
            filter=TolokaFilterCondition.deserialize(self.filter) if self.filter else None,
        )


@message
@dataclass
class HumanFilter(ProtobufSerializer[worker.HumanFilter]):
    filter: OneOf_ = one_of(
        expert=part(ExpertFilter, 1),
        worker=part(WorkerFilter, 2),
        custom_worker=part(CustomWorkerFilter, 3),
    )

    @staticmethod
    def serialize(obj: worker.HumanFilter) -> 'HumanFilter':
        f = HumanFilter()
        if isinstance(obj, worker.ExpertFilter):
            f.filter.expert = ExpertFilter.serialize(obj)
        elif isinstance(obj, worker.WorkerFilter):
            f.filter.worker = WorkerFilter.serialize(obj)
        elif isinstance(obj, worker.CustomWorkerFilter):
            f.filter.custom_worker = CustomWorkerFilter.serialize(obj)
        else:
            raise ValueError(f'unexpected filter type: {type(obj)}')
        return f

    def deserialize(self) -> worker.HumanFilter:
        return deserialize_one_of_field(self.filter)


@message
@dataclass
class Human(ProtobufSerializer[worker.Human]):
    user_id: str = field(1)
    assignment_id: str = field(2)

    @staticmethod
    def serialize(obj: worker.Human) -> 'Human':
        return Human(user_id=obj.user_id, assignment_id=obj.assignment_id)

    def deserialize(self) -> worker.Human:
        fake_assignment = toloka.Assignment(id=self.assignment_id, user_id=self.user_id)  # TODO: ugly hack
        return worker.Human(fake_assignment)


@message
@dataclass
class Model(ProtobufSerializer[worker.Model]):
    name: str = field(1)
    func: str = field(2)

    @staticmethod
    def serialize(obj: worker.Model) -> 'Model':
        return Model(name=obj.name, func=func_registry[obj.func])

    def deserialize(self) -> worker.Model:
        return worker.Model(name=self.name, func=load_func(self.func))


@message
@dataclass
class Worker(ProtobufSerializer[worker.Worker]):
    worker: OneOf_ = one_of(
        human=part(Human, 1),
        model=part(Model, 2),
    )

    @staticmethod
    def serialize(obj: T) -> 'Worker':
        w = Worker()
        if isinstance(obj, worker.Human):
            w.worker.human = Human.serialize(obj)
        elif isinstance(obj, worker.Model):
            w.worker.model = Model.serialize(obj)
        else:
            raise ValueError(f'unexpected worker type: {type(obj)}')
        return w

    def deserialize(self) -> T:
        return deserialize_one_of_field(self.worker)



from .base import *  # noqa
from .classification import *  # noqa
from .classification_loop import *  # noqa
from .client import *  # noqa
from .common import *  # noqa
from .control import *  # noqa
from .evaluation import *  # noqa
from .mapping import *  # noqa
from .pricing import *  # noqa
from .toloka import *  # noqa
from .worker import *  # noqa



from dataclasses import dataclass
from typing import List, Optional, Type

from pure_protobuf.dataclasses_ import field, message, one_of, part

from ... import classification
from .base import Label, StrEnum
from .common import ProtobufSerializer
from .worker import Worker


@message
@dataclass
class LabelProba(ProtobufSerializer[classification.LabelProba]):
    label: Label = field(1)
    proba: float = field(2)

    @staticmethod
    def serialize(obj: classification.LabelProba) -> 'LabelProba':
        label, proba = obj
        return LabelProba(label=Label.serialize(label), proba=proba)

    def deserialize(self) -> classification.LabelProba:
        return self.label.deserialize(), self.proba


@message
@dataclass
class TaskLabelsProbas(ProtobufSerializer[classification.TaskLabelsProbas]):
    items: List[LabelProba] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: classification.TaskLabelsProbas) -> 'TaskLabelsProbas':
        return TaskLabelsProbas(items=[LabelProba.serialize((label, proba)) for label, proba in obj.items()])

    def deserialize(self) -> classification.TaskLabelsProbas:
        return {item.label.deserialize(): item.proba for item in self.items}


@message
class AggregationAlgorithm(StrEnum[classification.AggregationAlgorithm]):
    @staticmethod
    def enum_cls() -> Type[classification.AggregationAlgorithm]:
        return classification.AggregationAlgorithm


@message
@dataclass
class WorkerLabel(ProtobufSerializer[classification.WorkerLabel]):
    label: Label = field(1)
    worker: Worker = field(2)

    @staticmethod
    def serialize(obj: classification.WorkerLabel) -> 'WorkerLabel':
        label, worker = obj
        return WorkerLabel(label=Label.serialize(label), worker=Worker.serialize(worker))

    def deserialize(self) -> classification.WorkerLabel:
        return self.label.deserialize(), self.worker.deserialize()


@message
@dataclass
class WorkerWeightsItem:
    worker_id: str = field(1)
    weight: float = field(2)


@message
@dataclass
class WorkerWeights(ProtobufSerializer[classification.WorkerWeights]):
    entries: List[WorkerWeightsItem] = field(1)

    @staticmethod
    def serialize(obj: classification.WorkerWeights) -> 'WorkerWeights':
        return WorkerWeights([WorkerWeightsItem(worker_id, weight) for worker_id, weight in obj.items()])

    def deserialize(self) -> classification.WorkerWeights:
        return {entry.worker_id: entry.weight for entry in self.entries}


@message
@dataclass
class WorkerLabels:
    labels: List[WorkerLabel] = field(1, default_factory=list)


worker_labels_empty = WorkerLabels(labels=[])


@message
@dataclass
class ResultsItem:
    probas: Optional[TaskLabelsProbas] = field(1, default=None)
    labels: WorkerLabels = field(2, default=worker_labels_empty)


@message
@dataclass
class Results(ProtobufSerializer[classification.Results]):
    items: List[ResultsItem] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: classification.Results) -> 'Results':
        return Results(
            [
                ResultsItem(
                    probas=TaskLabelsProbas.serialize(probas) if probas else None,
                    labels=WorkerLabels([WorkerLabel.serialize(label) for label in labels]),
                )
                for probas, labels in obj
            ]
        )

    def deserialize(self) -> classification.Results:
        return [
            (item.probas.deserialize() if item.probas else None, [label.deserialize() for label in item.labels.labels])
            for item in self.items
        ]



from random import choice
import threading
from typing import List, Tuple, Optional

import toloka.client as toloka

from .. import base, classification, mapping, task_spec as spec

from .launch import Params, launch
from .relaunch import get_swaps


# In a side-by-side comparison, client has two variants of objects, list of A's and list of B's, and want to know which
# variant, A or B, is better according to his criteria. By default, we place A's to the left, but to avoid workers bias
# of left/right variants of task UI or different exploits, we always randomly swap each pair of variants before creating
# tasks for workers. Overall process looks like:
# - take A's and B's (client input)
# - random swap, create L's and R's for tasks (remember swaps)
# - labeling
# - get results as L's and R's
# - convert L's and R's back to A's and B's (using remembered swaps)
# - return results for each pair in A's and B's terms (client output)
def launch_sbs(
    task_spec: spec.PreparedTaskSpec,
    params: Params,
    input_objects: List[mapping.Objects],
    control_objects: List[mapping.TaskSingleSolution],  # assisted by us
    client: toloka.TolokaClient,
    interactive: bool = False,
    interrupted: Optional[threading.Event] = None,
    name: Optional[str] = None,
) -> Tuple[classification.Results, Optional[classification.WorkerWeights]]:
    assert isinstance(task_spec.function, base.SbSFunction)

    # [hint, option_a, option_b] = [choice_a | choice_b]
    # swap: option_a <-> option_b, choice_a <-> choice_b

    h_cnt = len(task_spec.function.get_hints())
    i_cnt = len(task_spec.function.get_inputs())
    a_fr = h_cnt
    b_fr = h_cnt + i_cnt
    b_to = h_cnt + i_cnt * 2

    def swap_task(task: mapping.Objects) -> mapping.Objects:
        result = list(task)
        result[a_fr:b_fr], result[b_fr:b_to] = result[b_fr:b_to], result[a_fr:b_fr]
        return tuple(result)

    def swap_task_solution(task_solution: mapping.TaskSingleSolution) -> mapping.TaskSingleSolution:
        task, solution = task_solution
        return swap_task(task), (solution[0].swap(),) + solution[1:]

    input_objects_swapped = []
    swaps = get_swaps(name, input_objects)
    for swap, task_input_objects in zip(swaps, input_objects):
        swaps.append(swap)
        if swap:
            task_input_objects = swap_task(task_input_objects)
        input_objects_swapped.append(task_input_objects)

    control_objects_swapped = []
    for control_object in control_objects:
        swap = choice((True, False))
        if swap:
            control_object = swap_task_solution(control_object)
        control_objects_swapped.append(control_object)

    results_swapped, worker_weights = launch(
        task_spec=task_spec,
        params=params,
        input_objects=input_objects_swapped,
        control_objects=control_objects_swapped,
        client=client,
        interactive=interactive,
        interrupted=interrupted,
        name=name,
    )

    results = []
    for (labels_probas, worker_labels), swap in zip(results_swapped, swaps):
        assert labels_probas is not None
        if swap:
            labels_probas = {label.swap(): proba for label, proba in labels_probas.items()}
            worker_labels = [(label.swap(), worker) for label, worker in worker_labels]
        results.append((labels_probas, worker_labels))
    return results, worker_weights



from dataclasses import dataclass, field

import json
import toloka.client as toloka

from . import base, mapping, project


@dataclass
class PreparedTaskSpec:
    task_spec: base.TaskSpec
    lang: str
    scenario: project.Scenario = project.Scenario.DEFAULT

    view: toloka.project.view_spec.ViewSpec = field(init=False)
    task_mapping: mapping.TaskMapping = field(init=False)
    dumped_view: str = field(init=False)

    def __post_init__(self):
        if isinstance(self, AnnotationTaskSpec):
            builder = project.AnnotationBuilder
        else:
            builder = {
                base.ClassificationFunction: project.ClassificationBuilder,
                base.SbSFunction: project.SbSBuilder,
                base.AnnotationFunction: project.AnnotationCheckBuilder,
            }[type(self.function)]

        self.view, self.task_mapping = builder(task_function=self.function, lang=self.lang).get_scenario(
            self.scenario, self.for_preview()
        )
        self.dump_view()

    def dump_view(self):
        self.dumped_view = json.dumps(json.loads(self.view.unstructure()['config']), ensure_ascii=False)

    @property
    def id(self) -> str:
        return self.task_spec.id

    @property
    def function(self) -> base.TaskFunction:
        return self.task_spec.function

    @property
    def name(self) -> base.LocalizedString:
        if self.scenario == project.Scenario.DEFAULT:
            return self.task_spec.name
        return self.task_spec.name + ' (' + self.scenario.project_name_suffix + ')'

    @property
    def description(self) -> base.LocalizedString:
        return self.task_spec.description

    @property
    def instruction(self) -> base.LocalizedString:
        return self.task_spec.instruction

    def log_message(self) -> str:
        scenario = '' if self.scenario == project.Scenario.DEFAULT else f' to {self.scenario.name}'
        return f'task "{self.task_spec.id}" in {self.lang}{scenario}'

    def for_preview(self) -> bool:
        return False

    def overload_view(self, view: toloka.project.view_spec.ViewSpec):
        self.view = view
        self.dump_view()


@dataclass
class PreviewTaskSpec(PreparedTaskSpec):
    def for_preview(self) -> bool:
        return True


@dataclass
class AnnotationTaskSpec(PreparedTaskSpec):
    check: PreparedTaskSpec = field(init=False)

    def __post_init__(self):
        super().__post_init__()
        check_spec = base.TaskSpec(
            id=f'{self.task_spec.id}_check',
            function=self.task_spec.function,
            name=self.task_spec.name + ' – ' + project.CHECK,
            description=self.description + ' – ' + project.CHECK,
            instruction=self.instruction,
        )
        self.check = PreparedTaskSpec(check_spec, self.lang, self.scenario)


@dataclass
class AnnotationPreviewTaskSpec(PreviewTaskSpec, AnnotationTaskSpec):
    def __post_init__(self):
        super().__post_init__()
        # we don't use it for previews directly, but we need transformed task mapping in check spec here
        self.check = PreviewTaskSpec(self.check.task_spec, self.lang, self.scenario)



from datetime import timedelta
import pytest

from toloka.client.assignment import Assignment
from toloka.client.user_restriction import UserRestriction

from crowdom import control


class TestPredicate:
    def test_different_types_in_expression(self):
        with pytest.raises(AssertionError) as e:
            control.PredicateExpression(
                boolean_operator=control.BooleanOperator.AND,
                predicates=[
                    control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('>')),
                    control.AssignmentDurationPredicate(threshold=0.2, comparison=control.ComparisonType('>')),
                ],
            )

        assert (
            str(e.value) == f'All predicates should be of same type. Expected {control.AssignmentAccuracyPredicate}, '
            f'got {control.AssignmentDurationPredicate}'
        )

    def test_empty_list(self):
        with pytest.raises(AssertionError) as e:
            control.PredicateExpression(boolean_operator=control.BooleanOperator.AND, predicates=[])
        assert str(e.value) == 'List of predicates should be non-empty'

    def test_nested_expression(self):
        with pytest.raises(AssertionError) as e:
            control.PredicateExpression(
                boolean_operator=control.BooleanOperator.AND,
                predicates=[
                    control.PredicateExpression(
                        boolean_operator=control.BooleanOperator.AND,
                        predicates=[
                            control.AssignmentDurationPredicate(
                                threshold=0.5,
                                comparison=control.ComparisonType('<='),
                            ),
                            control.AssignmentDurationPredicate(
                                threshold=0.2,
                                comparison=control.ComparisonType('>'),
                            ),
                        ],
                    ),
                    control.PredicateExpression(
                        boolean_operator=control.BooleanOperator.AND,
                        predicates=[
                            control.AssignmentDurationPredicate(
                                threshold=0.5,
                                comparison=control.ComparisonType('<='),
                            ),
                            control.AssignmentDurationPredicate(
                                threshold=0.2,
                                comparison=control.ComparisonType('>'),
                            ),
                        ],
                    ),
                ],
            )

        assert str(e.value) == 'Nested expressions are not allowed'


class TestRuleBuilder:
    def test_static_reward(self):
        expected_rules = [
            control.Rule(
                predicate=control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('>=')),
                action=control.SetAssignmentStatus(status=Assignment.ACCEPTED),
            ),
            control.Rule(
                predicate=control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('<')),
                action=control.SetAssignmentStatus(status=Assignment.REJECTED),
            ),
        ]

        rules = control.RuleBuilder().add_static_reward(threshold=0.7).build()

        assert rules == expected_rules

    def test_dynamic_reward_single_split(self):
        expected_rules = [
            control.Rule(
                predicate=control.AssignmentAccuracyPredicate(threshold=0.1, comparison=control.ComparisonType('>=')),
                action=control.SetAssignmentStatus(status=Assignment.ACCEPTED),
            ),
            control.Rule(
                predicate=control.AssignmentAccuracyPredicate(threshold=0.1, comparison=control.ComparisonType('<')),
                action=control.SetAssignmentStatus(status=Assignment.REJECTED),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('>='))
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.03),
            ),
        ]

        rules = (
            control.RuleBuilder()
            .add_dynamic_reward(
                min_accuracy_for_accept=0.1,
                min_accuracy_for_bonus=0.7,
                min_bonus_amount_usd=0.03,
                max_bonus_amount_usd=0.03,
                bonus_granularity_num=1,
            )
            .build()
        )

        assert rules == expected_rules

    def test_dynamic_reward_default_min_bonus_accuracy(self):
        expected_rules = [
            control.Rule(
                predicate=control.AssignmentAccuracyPredicate(threshold=0.0, comparison=control.ComparisonType('>=')),
                action=control.SetAssignmentStatus(status=Assignment.ACCEPTED),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.0, comparison=control.ComparisonType('>=')),
                        control.AssignmentAccuracyPredicate(threshold=0.25, comparison=control.ComparisonType('<')),
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.03),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.25, comparison=control.ComparisonType('>=')),
                        control.AssignmentAccuracyPredicate(threshold=0.5, comparison=control.ComparisonType('<')),
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.04),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.5, comparison=control.ComparisonType('>=')),
                        control.AssignmentAccuracyPredicate(threshold=0.75, comparison=control.ComparisonType('<')),
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.05),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.75, comparison=control.ComparisonType('>='))
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.06),
            ),
        ]

        rules = (
            control.RuleBuilder()
            .add_dynamic_reward(
                min_bonus_amount_usd=0.03,
                max_bonus_amount_usd=0.06,
                bonus_granularity_num=4,
            )
            .build()
        )

        assert rules == expected_rules

    def test_dynamic_reward_granular(self):
        expected_rules = [
            control.Rule(
                predicate=control.AssignmentAccuracyPredicate(threshold=0.0, comparison=control.ComparisonType('>=')),
                action=control.SetAssignmentStatus(status=Assignment.ACCEPTED),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.6, comparison=control.ComparisonType('>=')),
                        control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('<')),
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.03),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('>=')),
                        control.AssignmentAccuracyPredicate(threshold=0.8, comparison=control.ComparisonType('<')),
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.04),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.8, comparison=control.ComparisonType('>=')),
                        control.AssignmentAccuracyPredicate(threshold=0.9, comparison=control.ComparisonType('<')),
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.05),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator('and'),
                    predicates=[
                        control.AssignmentAccuracyPredicate(threshold=0.9, comparison=control.ComparisonType('>='))
                    ],
                ),
                action=control.GiveBonusToUser(amount_usd=0.06),
            ),
        ]

        rules = (
            control.RuleBuilder()
            .add_dynamic_reward(
                min_accuracy_for_bonus=0.6,
                min_bonus_amount_usd=0.03,
                max_bonus_amount_usd=0.06,
                bonus_granularity_num=4,
            )
            .build()
        )

        assert rules == expected_rules

    def test_only_one_reward_type(self):
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_static_reward(0.5).add_dynamic_reward(
                min_bonus_amount_usd=0.01, max_bonus_amount_usd=0.1
            )
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_dynamic_reward(
                min_bonus_amount_usd=0.01, max_bonus_amount_usd=0.1
            ).add_static_reward(0.5)

    def test_static_reward_incorrect_thresholds(self):
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_static_reward(1.5)
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_static_reward(-1.5)

    def test_dynamic_reward_incorrect_thresholds(self):
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_dynamic_reward(min_bonus_amount_usd=1.0, max_bonus_amount_usd=0)
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_dynamic_reward(
                min_bonus_amount_usd=0.001,
                max_bonus_amount_usd=0.1,
                min_accuracy_for_bonus=0.1,
                min_accuracy_for_accept=1.0,
            )
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_dynamic_reward(
                min_bonus_amount_usd=0.01,
                max_bonus_amount_usd=0.04,
                min_accuracy_for_bonus=0.1,
                min_accuracy_for_accept=1.0,
                bonus_granularity_num=5,
            )
        with pytest.raises(AssertionError):
            control.RuleBuilder().add_dynamic_reward(
                min_bonus_amount_usd=0.0,
                max_bonus_amount_usd=0.0,
                min_accuracy_for_bonus=1.5,
                min_accuracy_for_accept=1.0,
            )

    def test_speed_control(self):
        expected_rules = [
            control.Rule(
                control.AssignmentDurationPredicate(threshold=0.2, comparison=control.ComparisonType('<=')),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL, private_comment='Fast submits, <= 0.2', duration=timedelta(days=1)
                ),
            ),
            control.Rule(
                control.AssignmentDurationPredicate(threshold=0.2, comparison=control.ComparisonType('<=')),
                action=control.SetAssignmentStatus(status=Assignment.REJECTED),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator.AND,
                    predicates=[
                        control.AssignmentDurationPredicate(
                            threshold=0.5,
                            comparison=control.ComparisonType('<='),
                        ),
                        control.AssignmentDurationPredicate(
                            threshold=0.2,
                            comparison=control.ComparisonType('>'),
                        ),
                    ],
                ),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL,
                    private_comment='Fast submits, 0.2 < time <= 0.5',
                    duration=timedelta(hours=8),
                ),
            ),
        ]

        rules = control.RuleBuilder().add_speed_control(ratio_rand=0.2, ratio_poor=0.5).rules

        assert rules == expected_rules

    def test_complex_speed_control_no_blocks(self):
        rules = control.RuleBuilder().add_complex_speed_control(speed_blocks=[]).rules
        assert rules == []

    def test_complex_speed_control_one_block(self):
        expected_rules = [
            control.Rule(
                control.AssignmentDurationPredicate(threshold=0.2, comparison=control.ComparisonType('<=')),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL, private_comment='Fast submits, <= 0.2', duration=timedelta(days=1)
                ),
            ),
        ]

        rules = (
            control.RuleBuilder()
            .add_complex_speed_control(speed_blocks=[control.BlockTimePicker(0.2, '1d', False)])
            .rules
        )

        assert rules == expected_rules

    def test_complex_speed_control_unordered_block(self):
        expected_rules = [
            control.Rule(
                control.AssignmentDurationPredicate(threshold=0.1, comparison=control.ComparisonType('<=')),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL, private_comment='Fast submits, <= 0.1', duration=timedelta(days=2)
                ),
            ),
            control.Rule(
                control.AssignmentDurationPredicate(threshold=0.1, comparison=control.ComparisonType('<=')),
                action=control.SetAssignmentStatus(status=Assignment.REJECTED),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator.AND,
                    predicates=[
                        control.AssignmentDurationPredicate(
                            threshold=0.2,
                            comparison=control.ComparisonType('<='),
                        ),
                        control.AssignmentDurationPredicate(
                            threshold=0.1,
                            comparison=control.ComparisonType('>'),
                        ),
                    ],
                ),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL,
                    private_comment='Fast submits, 0.1 < time <= 0.2',
                    duration=timedelta(days=1),
                ),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator.AND,
                    predicates=[
                        control.AssignmentDurationPredicate(
                            threshold=0.5,
                            comparison=control.ComparisonType('<='),
                        ),
                        control.AssignmentDurationPredicate(
                            threshold=0.2,
                            comparison=control.ComparisonType('>'),
                        ),
                    ],
                ),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL,
                    private_comment='Fast submits, 0.2 < time <= 0.5',
                    duration=timedelta(minutes=30),
                ),
            ),
        ]

        rules = (
            control.RuleBuilder()
            .add_complex_speed_control(
                speed_blocks=[
                    control.BlockTimePicker(0.2, '1d', False),
                    control.BlockTimePicker(0.1, '2d', True),
                    control.BlockTimePicker(0.5, '30m', False),
                ],
            )
            .rules
        )

        assert rules == expected_rules

    def test_control_task_control_no_hard_block(self):
        rules = (
            control.RuleBuilder()
            .add_control_task_control(
                control_task_count=10,
                control_task_correct_count_for_hard_block=0,
                control_task_correct_count_for_soft_block=2,
            )
            .rules
        )
        expected_rules = [
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator.AND,
                    predicates=[
                        control.AssignmentAccuracyPredicate(
                            threshold=0.0, comparison=control.ComparisonType.GREATER_OR_EQUAL
                        ),
                        control.AssignmentAccuracyPredicate(threshold=0.2, comparison=control.ComparisonType.LESS),
                    ],
                ),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL,
                    private_comment='Control tasks: [0, 2) done correctly',
                    duration=timedelta(hours=1),
                ),
            ),
        ]
        assert rules == expected_rules

    def test_control_task_control_with_hard_block(self):
        rules = (
            control.RuleBuilder()
            .add_control_task_control(
                control_task_count=10,
                control_task_correct_count_for_hard_block=2,
                control_task_correct_count_for_soft_block=5,
            )
            .rules
        )
        expected_rules = [
            control.Rule(
                predicate=control.AssignmentAccuracyPredicate(threshold=0.2, comparison=control.ComparisonType.LESS),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL,
                    private_comment='Control tasks: [0, 2) done correctly',
                    duration=timedelta(hours=8),
                ),
            ),
            control.Rule(
                predicate=control.PredicateExpression(
                    boolean_operator=control.BooleanOperator.AND,
                    predicates=[
                        control.AssignmentAccuracyPredicate(
                            threshold=0.2, comparison=control.ComparisonType.GREATER_OR_EQUAL
                        ),
                        control.AssignmentAccuracyPredicate(threshold=0.5, comparison=control.ComparisonType.LESS),
                    ],
                ),
                action=control.BlockUser(
                    scope=UserRestriction.Scope.POOL,
                    private_comment='Control tasks: [2, 5) done correctly',
                    duration=timedelta(hours=1),
                ),
            ),
        ]
        assert rules == expected_rules


class TestControl:
    def test_control_filter_rules_basic(self):
        accept_rule = control.Rule(
            predicate=control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('>=')),
            action=control.SetAssignmentStatus(status=Assignment.ACCEPTED),
        )
        reject_rule = control.Rule(
            predicate=control.AssignmentAccuracyPredicate(threshold=0.7, comparison=control.ComparisonType('<')),
            action=control.SetAssignmentStatus(status=Assignment.REJECTED),
        )
        block_rule = control.Rule(
            predicate=control.AssignmentAccuracyPredicate(threshold=0.3, comparison=control.ComparisonType('<')),
            action=control.BlockUser(private_comment='Bad performance', scope=UserRestriction.PROJECT),
        )
        speed_block_rule = control.Rule(
            predicate=control.AssignmentDurationPredicate(threshold=0.5, comparison=control.ComparisonType('<=')),
            action=control.BlockUser(private_comment='Fast submits', scope=UserRestriction.PROJECT),
        )
        bonus_rule = control.Rule(
            predicate=control.AssignmentAccuracyPredicate(threshold=0.9, comparison=control.ComparisonType('>')),
            action=control.GiveBonusToUser(amount_usd=0.03),
        )

        ctrl = control.Control(rules=[accept_rule, reject_rule, block_rule, speed_block_rule, bonus_rule])

        for predicate_type, action_type, rules in [
            (control.AssignmentAccuracyPredicate, control.SetAssignmentStatus, [accept_rule, reject_rule]),
            (control.AssignmentAccuracyPredicate, control.BlockUser, [block_rule]),
            (control.AssignmentAccuracyPredicate, control.GiveBonusToUser, [bonus_rule]),
        ]:
            assert ctrl.filter_rules(predicate_type=predicate_type, action_type=action_type) == rules

    def test_control_filter_rules_expression(self):
        block_rule = control.Rule(
            predicate=control.PredicateExpression(
                boolean_operator=control.BooleanOperator.AND,
                predicates=[
                    control.AssignmentAccuracyPredicate(threshold=0.3, comparison=control.ComparisonType('<')),
                    control.AssignmentAccuracyPredicate(threshold=0.1, comparison=control.ComparisonType('>=')),
                ],
            ),
            action=control.BlockUser(private_comment='Bad performance', scope=UserRestriction.PROJECT),
        )
        speed_block_rule = control.Rule(
            predicate=control.AssignmentDurationPredicate(threshold=0.5, comparison=control.ComparisonType('<=')),
            action=control.BlockUser(private_comment='Fast submits', scope=UserRestriction.PROJECT),
        )
        bonus_rule = control.Rule(
            predicate=control.PredicateExpression(
                boolean_operator=control.BooleanOperator.AND,
                predicates=[
                    control.AssignmentAccuracyPredicate(threshold=0.9, comparison=control.ComparisonType('>')),
                    control.AssignmentAccuracyPredicate(threshold=0.95, comparison=control.ComparisonType('<=')),
                ],
            ),
            action=control.GiveBonusToUser(amount_usd=0.03),
        )

        ctrl = control.Control(rules=[block_rule, speed_block_rule, bonus_rule])

        for predicate_type, action_type, rules in [
            (control.AssignmentAccuracyPredicate, control.BlockUser, [block_rule]),
            (control.AssignmentAccuracyPredicate, control.GiveBonusToUser, [bonus_rule]),
        ]:
            assert ctrl.filter_rules(predicate_type=predicate_type, action_type=action_type) == rules



from enum import Enum
import json
from typing import List, Optional, Type, Tuple, ClassVar, Dict, Union, Literal

import numpy as np
import pandas as pd

from .. import (
    base,
    classification,
    contrib,
    feedback_loop,
    mapping,
    objects,
    project,
    task_spec as spec,
)

from ..base import TaskSpec
from ..worker import NoWorker
from ..utils import DecimalEncoder


def get_spec_cls(task_function: base.TaskFunction, for_preview: bool = False) -> Type[spec.PreparedTaskSpec]:
    if type(task_function) in [base.ClassificationFunction, base.SbSFunction]:
        return spec.PreviewTaskSpec if for_preview else spec.PreparedTaskSpec
    elif isinstance(task_function, base.AnnotationFunction):
        return spec.AnnotationPreviewTaskSpec if for_preview else spec.AnnotationTaskSpec
    else:
        raise ValueError(f'Unknown task function: {task_function}')


def get_spec_cls_by_task_spec(task_spec: spec.PreparedTaskSpec) -> Type[spec.PreparedTaskSpec]:
    if isinstance(task_spec, spec.AnnotationTaskSpec):
        return spec.AnnotationPreviewTaskSpec
    return spec.PreviewTaskSpec


# template builder saves config and input fields in encoded way to URL, but it doesn't save output fields and state of
# UI, so we can create preview just for task (without solution), and there will be unnecessary config block
class TaskPreview:
    def __init__(
        self,
        input_objects: mapping.Objects,
        *,
        task_spec: Optional[spec.PreparedTaskSpec] = None,
        task_function: Optional[base.TaskFunction] = None,
        lang: Optional[str] = None,
        output_objects: Optional[mapping.Objects] = None,
    ):
        assert (
            task_spec is not None or task_function is not None and lang is not None
        ), 'Provide either task_spec or both task_function and language'
        for_preview = output_objects is not None
        if task_spec is None:
            task_spec = get_spec_cls(task_function, for_preview)(
                task_spec=TaskSpec(
                    id=None,
                    function=task_function,
                    name=base.EMPTY_STRING,
                    description=base.EMPTY_STRING,
                    instruction=None,
                ),
                lang=lang,
            )
        elif for_preview:
            task_spec = get_spec_cls_by_task_spec(task_spec)(
                task_spec=task_spec.task_spec,
                lang=task_spec.lang,
                scenario=task_spec.scenario,
            )
        self.url = self.get_url(input_objects, task_spec, output_objects)

    def get_link(self) -> str:
        return f'<a href="{self.url}" target="_blank" rel="noopener noreferrer">task preview</a>'

    def display_link(self):
        from IPython.display import HTML

        return HTML(self.get_link())

    def display(self):
        from IPython.display import HTML

        return HTML(
            f'<iframe src="{self.url}" onload="javascript:(function(o)'
            '{o.style.height=o.contentWindow.document.body.scrollHeight+"px";}(this));" '
            'style="height:800px;width:100%;border:none;overflow:hidden;"/>'
        )

    @classmethod
    def get_url(
        cls,
        input_objects: mapping.Objects,
        task_spec: spec.PreparedTaskSpec,
        output_objects: Optional[mapping.Objects] = None,
    ) -> str:
        objects_count = len(input_objects)
        if output_objects is not None:
            # for results preview we display output types as inputs, due to tb limitations
            objects_count += len(output_objects)

        if isinstance(task_spec, spec.AnnotationTaskSpec):
            if objects_count == len(task_spec.check.task_mapping.input_mapping):
                task_spec = task_spec.check
            else:
                assert objects_count == len(
                    task_spec.task_mapping.input_mapping
                ), 'incorrect number of input objects supplied'
        config, task_mapping = task_spec.dumped_view, task_spec.task_mapping
        preview_objects = input_objects
        if output_objects is not None:
            preview_objects = np.concatenate([input_objects, output_objects])
        input_values: dict = task_mapping.to_task(preview_objects).input_values

        dumped = json.dumps(
            {'config': config, 'input': json.dumps(input_values, ensure_ascii=False, cls=DecimalEncoder)},
            ensure_ascii=False,
        )
        compressed = contrib.LZString.compressToEncodedURIComponent(dumped)
        return f'https://tb.yandex.net/editor?config={compressed}'


class Results:
    row_input_objects: np.ndarray
    row_output_objects: np.ndarray

    task_spec: spec.PreparedTaskSpec

    WORKER_FIELD: ClassVar[str] = 'worker'
    PREVIEW_FIELD: ClassVar[str] = 'preview'

    def _add_task_previews(self, df: pd.DataFrame) -> pd.DataFrame:
        urls = [
            TaskPreview(task_input_objects, output_objects=task_output_objects, task_spec=self.task_spec).get_link()
            for task_input_objects, task_output_objects in zip(
                self.row_input_objects[df.index], self.row_output_objects[df.index]
            )
        ]
        df_copy = df.copy()
        df_copy[self.PREVIEW_FIELD] = urls
        return df_copy

    def html_with_task_previews(self, df: pd.DataFrame):
        from IPython.display import HTML

        return HTML(self._add_task_previews(df).to_html(escape=False))

    @staticmethod
    def deduplicate(df: pd.DataFrame) -> pd.DataFrame:
        return df.loc[df.astype(str).drop_duplicates().index]


class LabelingResults(Results):
    df: pd.DataFrame
    has_worker_weights: bool
    RESULT_FIELD: ClassVar[str] = 'result'
    CONFIDENCE_FIELD: ClassVar[str] = 'confidence'


# currently we only have 1 use case for combined classes, and producing M * N proba items is excessive,
# because all possible answers for each question are pooled from same 5 options
# thus, currently we revert generated answers to their original form inside Results
def get_class_fields(cls_type: Type[base.Class]) -> List[Tuple[str, str]]:
    if issubclass(cls_type, objects.CombinedAnswer):
        cls_values = list(set(cls.get_original_answer().value for cls in cls_type.possible_instances()))
    else:
        cls_values = [cls.value for cls in cls_type.possible_instances()]

    return [(cls_value, f'proba_{cls_value}') for cls_value in cls_values]


def get_labels_probas(
    labels_probas: Optional[classification.TaskLabelsProbas], cls_type: Type[base.Class]
) -> Optional[classification.TaskLabelsProbas]:
    if labels_probas is None:
        return None
    if not issubclass(cls_type, objects.CombinedAnswer):
        return {key.value: value for key, value in labels_probas.items()}

    return {key.get_original_answer().value: value for key, value in labels_probas.items()}


def get_label(raw_label: Optional[base.Label]) -> Union[base.Label, Literal[np.nan]]:
    if raw_label is None:
        return np.nan
    if not isinstance(raw_label, objects.CombinedAnswer):
        return raw_label.value

    return raw_label.get_original_answer().value


class ClassificationResults(LabelingResults):
    WORKER_WEIGHT_FIELD: ClassVar[str] = 'worker_weight'
    LABEL_FIELD: ClassVar[str] = 'label'
    OVERLAP_FIELD: ClassVar[str] = 'overlap'

    def __init__(
        self,
        input_objects: List[mapping.Objects],
        results: List[Tuple[classification.TaskLabelsProbas, List[classification.WorkerLabel]]],
        task_spec: spec.PreparedTaskSpec,
        worker_weights: Optional[classification.WorkerWeights] = None,
    ):
        self.task_spec = task_spec

        # we will index these input objects by dataframe row indexes
        row_input_objects, row_output_objects = [], []

        input_fields = []
        for obj_mapping in task_spec.task_mapping.input_mapping:
            for _, task_field in obj_mapping.obj_task_fields:
                input_fields.append(task_field)

        self.input_fields = input_fields

        cls_type: Type[base.Class] = task_spec.task_mapping.output_mapping[0].obj_meta.type
        cls_fields = get_class_fields(cls_type)

        self.proba_fields = [cls_field for _, cls_field in cls_fields]
        self.has_worker_weights = worker_weights is not None
        rows = []
        for task_input_objects, task_result in zip(input_objects, results):
            task_row = task_spec.task_mapping.toloka_values(task_input_objects)
            del task_row[mapping.TASK_ID_FIELD]

            labels_probas, raw_labels = task_result
            labels_probas = get_labels_probas(labels_probas, cls_type)
            overlap = len(raw_labels)

            most_probable_result = classification.get_most_probable_label(labels_probas)
            if most_probable_result is not None:
                task_result, proba = most_probable_result
            else:
                assert labels_probas is None
                assert not raw_labels

                # todo: maybe confidence should also be None
                task_result, proba, labels_probas = None, 0.0, {}
                raw_labels = [(None, NoWorker())]
                if self.has_worker_weights:
                    worker_weights[None] = 0.0

            task_row[self.RESULT_FIELD] = np.nan if task_result is None else task_result
            task_row[self.CONFIDENCE_FIELD] = proba
            task_row[self.OVERLAP_FIELD] = overlap
            for cls_value, cls_field in cls_fields:
                proba = labels_probas.get(cls_value, 0.0)
                task_row[cls_field] = proba

            for raw_label, worker in raw_labels:
                worker_row = dict(task_row)
                worker_row[self.LABEL_FIELD] = get_label(raw_label)
                worker_row[self.WORKER_FIELD] = worker.id
                if self.has_worker_weights:
                    worker_row[self.WORKER_WEIGHT_FIELD] = worker_weights[worker.id]
                rows.append(worker_row)
                row_input_objects.append(task_input_objects)
                row_output_objects.append((raw_label,))

        self.df = pd.DataFrame.from_dict(rows)
        self.row_input_objects = np.array(row_input_objects)
        self.row_output_objects = np.array(row_output_objects)

    def predict(self) -> pd.DataFrame:
        return self.deduplicate(
            self.df[self.input_fields + [self.RESULT_FIELD, self.CONFIDENCE_FIELD, self.OVERLAP_FIELD]]
        )

    def predict_proba(self) -> pd.DataFrame:
        return self.deduplicate(self.df[self.input_fields + self.proba_fields + [self.OVERLAP_FIELD]])

    def worker_labels(self) -> pd.DataFrame:
        worker_labels = self.df[
            self.input_fields
            + [self.LABEL_FIELD, self.WORKER_FIELD]
            + ([self.WORKER_WEIGHT_FIELD] if self.has_worker_weights else [])
        ]
        return worker_labels[worker_labels[self.WORKER_FIELD].notnull()]


class AnnotationResults(LabelingResults):
    EVAL_FIELD: ClassVar[str] = base.EVALUATION_TASK_FIELD
    ANNOTATION_WORKER_FIELD: ClassVar[str] = 'annotator'
    CHECK_WORKER_FIELD: ClassVar[str] = 'evaluator'
    WORKER_WEIGHT_FIELD: ClassVar[str] = 'evaluator_weight'
    CHECK_OVERLAP_FIELD: ClassVar[str] = 'evaluation_overlap'
    ANNOTATION_OVERLAP_FIELD: ClassVar[str] = 'annotation_overlap'

    def __init__(
        self,
        input_objects: List[mapping.Objects],
        results: List[List[feedback_loop.Solution]],
        task_spec: spec.AnnotationTaskSpec,
        worker_weights: Optional[classification.WorkerWeights] = None,
    ):
        self.task_spec = task_spec

        row_input_objects, row_output_objects = [], []

        input_fields = []
        for obj_mapping in task_spec.task_mapping.input_mapping:
            for _, task_field in obj_mapping.obj_task_fields:
                input_fields.append(task_field)

        self.input_fields = input_fields

        output_fields = []
        for obj_mapping in task_spec.task_mapping.output_mapping:
            for _, task_field in obj_mapping.obj_task_fields:
                output_fields.append(task_field)

        self.output_fields = output_fields

        self.has_worker_weights = worker_weights is not None

        rows = []
        for task_input_objects, solutions in zip(input_objects, results):
            inputs = task_spec.task_mapping.toloka_values(task_input_objects)
            del inputs[mapping.TASK_ID_FIELD]

            for i, solution in enumerate(solutions):
                outputs = task_spec.task_mapping.toloka_values(solution.solution, output=True)

                task_row = {**inputs, **outputs}
                task_row[self.ANNOTATION_WORKER_FIELD] = solution.worker.id

                task_row[self.RESULT_FIELD] = i == 0
                task_row[self.ANNOTATION_OVERLAP_FIELD] = len(solutions)
                if solution.evaluation is None:
                    task_row[self.CHECK_OVERLAP_FIELD] = 0
                    rows.append(task_row)
                    row_input_objects.append(task_input_objects)
                    row_output_objects.append(solution.solution)
                else:
                    assert solution.evaluation.worker_labels
                    task_row[self.CONFIDENCE_FIELD] = solution.evaluation.confidence
                    task_row[self.CHECK_OVERLAP_FIELD] = len(solution.evaluation.worker_labels)
                    for choice, worker in solution.evaluation.worker_labels:
                        worker_row = dict(task_row)
                        worker_row[self.EVAL_FIELD] = choice.ok
                        worker_row[self.CHECK_WORKER_FIELD] = worker.id
                        if self.has_worker_weights:
                            worker_row[self.WORKER_WEIGHT_FIELD] = worker_weights[worker.id]
                        rows.append(worker_row)
                        row_input_objects.append(task_input_objects)
                        row_output_objects.append(solution.solution)

        self.df = pd.DataFrame.from_dict(rows)
        self.row_input_objects = np.array(row_input_objects)
        self.row_output_objects = np.array(row_output_objects)

    def predict(self) -> pd.DataFrame:
        df = self.deduplicate(self.df[self.input_fields + self.output_fields + [self.RESULT_FIELD]])
        return df[df[self.RESULT_FIELD]].drop([self.RESULT_FIELD], axis=1)

    def predict_proba(self) -> pd.DataFrame:
        return self.deduplicate(self.df[self.input_fields + self.output_fields + [self.CONFIDENCE_FIELD]])

    def worker_labels(self) -> pd.DataFrame:
        return self.df.drop([self.RESULT_FIELD], axis=1)


class ExpertLabelingApplication(Enum):
    TRAINING = 'training'
    CONTROL_TASKS = 'control_tasks'
    ANNOTATION_CHECK_TRAINING = 'check_training'


class ExpertLabelingResults(Results):
    df: pd.DataFrame

    OK_FIELD: ClassVar[str] = '_ok'

    # todo implement this in ClassificationResults
    DURATION_FIELD: ClassVar[str] = 'duration'

    input_length: int
    output_lenght: int
    ok_index: int
    evaluation_index: Optional[int] = None
    comment_index: int
    output_names: List[str]

    def __init__(
        self,
        results: List[Tuple[Union[mapping.TaskSingleSolution, mapping.Objects], mapping.TaskMultipleSolutions]],
        task_spec: spec.PreparedTaskSpec,
        worker_id_to_name: Optional[Dict[str, str]] = None,
    ):
        self.task_spec = task_spec
        if worker_id_to_name is None:
            worker_id_to_name = {}
        # we will index these input objects by dataframe row indexes
        row_input_objects, row_output_object = [], []

        rows = []
        for _, (input_objects, solution) in results:
            if not solution:
                # expert labeling was interrupted, and this task doesn't have a solution
                continue
            [(output_objects, assignment)] = solution
            task_row = {
                **task_spec.task_mapping.toloka_values(input_objects, output=False),
                **task_spec.task_mapping.toloka_values(output_objects, output=True),
            }
            del task_row[mapping.TASK_ID_FIELD]
            task_row[self.WORKER_FIELD] = worker_id_to_name.get(assignment.user_id, assignment.user_id)
            task_row[self.DURATION_FIELD] = (assignment.submitted - assignment.created) / len(assignment.tasks)
            rows.append(task_row)
            row_input_objects.append(input_objects)
            row_output_object.append(output_objects)
        self.rows = rows
        self.df = pd.DataFrame.from_dict(rows)
        self.row_input_objects = np.array(row_input_objects)
        self.row_output_objects = np.array(row_output_object)
        function = task_spec.function

        default_spec = get_spec_cls(function)(task_spec=self.task_spec.task_spec, lang=self.task_spec.lang)

        self.input_length = len(default_spec.task_mapping.input_mapping)
        self.output_length = len(default_spec.task_mapping.output_mapping)
        ok_meta, comment_meta = project.Builder.EXPERT_LABELING_METAS

        self.output_names = [
            task_mapping.obj_task_fields[0][1] for task_mapping in self.task_spec.task_mapping.output_mapping
        ]
        self.ok_index = self.output_names.index(ok_meta.name)
        self.comment_index = self.output_names.index(comment_meta.name)
        if isinstance(function, base.AnnotationFunction):
            evaluation_name = function.get_evaluation_named_meta().name
            if evaluation_name in self.output_names:
                self.evaluation_index = self.output_names.index(evaluation_name)

    def get_results(self) -> pd.DataFrame:
        return self.df

    def get_accuracy(self):
        return self.df[self.OK_FIELD].mean()

    def get_solved_task_and_comment(
        self,
        task: mapping.Objects,
        solution: mapping.Objects,
        application: ExpertLabelingApplication,
    ) -> Tuple[mapping.TaskSingleSolution, objects.Text]:
        objs = task + solution
        if self.evaluation_index is None or application == ExpertLabelingApplication.TRAINING:
            inputs, outputs, comment = (
                objs[: self.input_length],
                objs[self.input_length : self.input_length + self.output_length],
                solution[self.comment_index],
            )
        else:
            inputs, outputs, comment = (
                objs[: self.input_length + self.output_length],
                (solution[self.evaluation_index],),
                solution[self.comment_index],
            )
        return (inputs, outputs), comment

    def suitable(self, solution: mapping.Objects, application: ExpertLabelingApplication) -> bool:
        if not solution[self.ok_index].ok:
            return False
        if application == ExpertLabelingApplication.CONTROL_TASKS:
            return True
        if solution[self.comment_index] is None or not solution[self.comment_index].text:
            return False
        if application == ExpertLabelingApplication.ANNOTATION_CHECK_TRAINING:
            return True
        if self.evaluation_index is not None:
            return solution[self.evaluation_index].ok
        return True

    def get_correct_objects(
        self,
        application: ExpertLabelingApplication = ExpertLabelingApplication.CONTROL_TASKS,
    ) -> Tuple[List[mapping.TaskSingleSolution], List[objects.Text]]:
        tasks_and_solutions = [
            (self.task_spec.task_mapping.from_task_values(row), self.task_spec.task_mapping.from_solution_values(row))
            for row in self.rows
        ]
        generated_objects = [
            self.get_solved_task_and_comment(task, solution, application)
            for task, solution in tasks_and_solutions
            if self.suitable(solution, application)
        ]
        solutions = [solution for solution, _ in generated_objects]
        texts = [text for _, text in generated_objects]
        return solutions, texts



from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import List, Optional, Type, Union

from pure_protobuf.dataclasses_ import field, message, one_of, part
from pure_protobuf.oneof import OneOf_
import toloka.client as toloka

from ... import control
from .base import StrEnum
from .common import ProtobufSerializer, deserialize_one_of_field


@message
class ComparisonType(StrEnum[control.ComparisonType]):
    @staticmethod
    def enum_cls() -> Type[control.ComparisonType]:
        return control.ComparisonType


@message
@dataclass
class PredicateValue(ProtobufSerializer[Union[float, timedelta]]):
    value: OneOf_ = one_of(
        float_=part(float, 1),
        timedelta=part(timedelta, 2),
    )

    @staticmethod
    def serialize(obj: Union[float, timedelta]) -> 'PredicateValue':
        value = PredicateValue()
        if isinstance(obj, timedelta):
            value.value.timedelta = obj
        else:
            value.value.float_ = obj
        return value

    def deserialize(self) -> Union[float, timedelta]:
        return deserialize_one_of_field(self.value)


@message
@dataclass
class ThresholdComparisonPredicate(ProtobufSerializer[control.ThresholdComparisonPredicate]):
    threshold: PredicateValue = field(1)
    comparison: ComparisonType = field(2)

    @staticmethod
    def serialize(obj: control.ThresholdComparisonPredicate) -> 'ThresholdComparisonPredicate':
        return ThresholdComparisonPredicate(
            threshold=PredicateValue.serialize(obj.threshold),
            comparison=ComparisonType.serialize(obj.comparison),
        )

    def deserialize(self) -> control.ThresholdComparisonPredicate:
        return control.ThresholdComparisonPredicate(
            threshold=self.threshold.deserialize(),
            comparison=self.comparison.deserialize(),
        )


@message
class BooleanOperator(StrEnum[control.BooleanOperator]):
    @staticmethod
    def enum_cls() -> Type[control.BooleanOperator]:
        return control.BooleanOperator


# better option is to pack predicate types to one_of, but it's not possible because expression predicate
# contains list of Predicates itself, but we can't pass class to one_of part using quotation syntax like 'Predicate'
@dataclass
class Predicate(ProtobufSerializer[control.Predicate]):
    assignment_accuracy: Optional[ThresholdComparisonPredicate] = field(1, default=None)
    assignment_duration: Optional[ThresholdComparisonPredicate] = field(2, default=None)
    always_true: bool = field(3, default=False)
    expression_boolean_operator: Optional[BooleanOperator] = field(4, default=None)
    expression_predicates: List['Predicate'] = field(5, default_factory=list)

    @staticmethod
    def serialize(obj: control.Predicate) -> 'Predicate':
        predicate = Predicate()
        if isinstance(obj, control.AssignmentAccuracyPredicate):
            predicate.assignment_accuracy = ThresholdComparisonPredicate.serialize(obj)
        elif isinstance(obj, control.AssignmentDurationPredicate):
            predicate.assignment_duration = ThresholdComparisonPredicate.serialize(obj)
        elif isinstance(obj, control.AlwaysTruePredicate):
            predicate.always_true = True
        elif isinstance(obj, control.PredicateExpression):
            predicate.expression_boolean_operator = BooleanOperator.serialize(obj.boolean_operator)
            predicate.expression_predicates = [Predicate.serialize(p) for p in obj.predicates]
        else:
            raise ValueError(f'unexpected predicate type: {type(obj)}')
        return predicate

    def deserialize(self) -> control.Predicate:
        threshold_proto = self.assignment_accuracy or self.assignment_duration
        if threshold_proto:
            threshold = ThresholdComparisonPredicate.deserialize(threshold_proto)
            type = (
                control.AssignmentAccuracyPredicate if self.assignment_accuracy else control.AssignmentDurationPredicate
            )
            return type(threshold=threshold.threshold, comparison=threshold.comparison)

        if self.always_true:
            return control.AlwaysTruePredicate()

        assert self.expression_boolean_operator and self.expression_predicates

        return control.PredicateExpression(
            boolean_operator=BooleanOperator.deserialize(self.expression_boolean_operator),
            predicates=[Predicate.deserialize(p) for p in self.expression_predicates],
        )


Predicate = message(Predicate)


@message
@dataclass
class BlockUser(ProtobufSerializer[control.BlockUser]):
    scope: str = field(1)
    private_comment: str = field(2)
    duration: Optional[timedelta] = field(3, default=None)

    @staticmethod
    def serialize(obj: control.BlockUser) -> 'BlockUser':
        return BlockUser(
            scope=obj.scope.value,
            private_comment=obj.private_comment,
            duration=obj.duration,
        )

    def deserialize(self) -> control.BlockUser:
        return control.BlockUser(
            scope=toloka.user_restriction.UserRestriction.Scope(self.scope),
            private_comment=self.private_comment,
            duration=self.duration,
        )


@message
@dataclass
class GiveBonusToUser(ProtobufSerializer[control.GiveBonusToUser]):
    amount_usd: float = field(1)

    @staticmethod
    def serialize(obj: control.GiveBonusToUser) -> 'GiveBonusToUser':
        return GiveBonusToUser(amount_usd=obj.amount_usd)

    def deserialize(self) -> control.GiveBonusToUser:
        return control.GiveBonusToUser(amount_usd=self.amount_usd)


@message
@dataclass
class SetAssignmentStatus(ProtobufSerializer[control.SetAssignmentStatus]):
    status: str = field(1)

    @staticmethod
    def serialize(obj: control.SetAssignmentStatus) -> 'SetAssignmentStatus':
        return SetAssignmentStatus(status=obj.status.value)

    def deserialize(self) -> control.SetAssignmentStatus:
        return control.SetAssignmentStatus(status=toloka.Assignment.Status(self.status))


@message
@dataclass
class Action(ProtobufSerializer[control.Action]):
    action: OneOf_ = one_of(
        block_user=part(BlockUser, 1),
        give_bonus_to_user=part(GiveBonusToUser, 2),
        set_assignment_status=part(SetAssignmentStatus, 3),
    )

    @staticmethod
    def serialize(obj: control.Action) -> 'Action':
        action = Action()
        if isinstance(obj, control.BlockUser):
            action.action.block_user = BlockUser.serialize(obj)
        elif isinstance(obj, control.GiveBonusToUser):
            action.action.give_bonus_to_user = GiveBonusToUser.serialize(obj)
        elif isinstance(obj, control.SetAssignmentStatus):
            action.action.set_assignment_status = SetAssignmentStatus.serialize(obj)
        else:
            raise ValueError(f'unexpected action type: {type(obj)}')
        return action

    def deserialize(self) -> control.Action:
        return deserialize_one_of_field(self.action)


@message
@dataclass
class Rule(ProtobufSerializer[control.Rule]):
    predicate: Predicate = field(1)
    action: Action = field(2)

    @staticmethod
    def serialize(obj: control.Rule) -> 'Rule':
        return Rule(predicate=Predicate.serialize(obj.predicate), action=Action.serialize(obj.action))

    def deserialize(self) -> control.Rule:
        return control.Rule(predicate=self.predicate.deserialize(), action=self.action.deserialize())


@message
@dataclass
class Control(ProtobufSerializer[control.Control]):
    rules: List[Rule] = field(1, default_factory=list)

    @staticmethod
    def serialize(obj: control.Control) -> 'Control':
        return Control([Rule.serialize(rule) for rule in obj.rules])

    def deserialize(self) -> control.Control:
        return control.Control([rule.deserialize() for rule in self.rules])



from typing import Dict

from crowdom import base, objects


def test_question_answers():
    class Dictor(objects.Question):
        SAME = 'same'
        DIFFERENT = 'different'
        OTHER = 'other'

        @classmethod
        def question_label(cls) -> Dict[str, str]:
            return {'EN': 'Is dictor the same for both audios?', 'RU': 'На записях звучит один и тот же голос?'}

        @classmethod
        def labels(cls) -> Dict['Dictor', Dict[str, str]]:
            return {
                cls.SAME: {'EN': 'same', 'RU': 'одинаковые'},
                cls.DIFFERENT: {'EN': 'different', 'RU': 'разные'},
                cls.OTHER: {'EN': 'other', 'RU': 'другое'},
            }

    class Noise(objects.Question):
        YES = 'yes'
        NO = 'no'

        @classmethod
        def question_label(cls) -> Dict[str, str]:
            return {'EN': 'Is there any noise in the audio?', 'RU': 'На записи присутствует шум?'}

    Question, Answer, question_answers = objects.get_combined_classes([Dictor, Noise])
    assert issubclass(Answer, objects.CombinedAnswer)
    assert issubclass(Answer, base.Class)

    assert len(Answer) == 5
    assert {answer.name: answer.value for answer in Answer} == {
        'DICTOR__DIFFERENT': 'dictor__different',
        'DICTOR__OTHER': 'dictor__other',
        'DICTOR__SAME': 'dictor__same',
        'NOISE__NO': 'noise__no',
        'NOISE__YES': 'noise__yes',
    }
    assert Answer.labels() == {
        Answer.DICTOR__DIFFERENT: {'EN': 'different', 'RU': 'разные'},
        Answer.DICTOR__OTHER: {'EN': 'other', 'RU': 'другое'},
        Answer.DICTOR__SAME: {'EN': 'same', 'RU': 'одинаковые'},
        Answer.NOISE__NO: {'EN': 'no'},
        Answer.NOISE__YES: {'EN': 'yes'},
    }

    assert Answer.DICTOR__DIFFERENT.get_original_answer() == Dictor.DIFFERENT
    assert Answer.DICTOR__OTHER.get_original_answer() == Dictor.OTHER
    assert Answer.DICTOR__SAME.get_original_answer() == Dictor.SAME
    assert Answer.NOISE__NO.get_original_answer() == Noise.NO
    assert Answer.NOISE__YES.get_original_answer() == Noise.YES

    assert issubclass(Question, base.Class)

    assert len(Question) == 2
    assert {Question.name: Question.value for Question in Question} == {
        'DICTOR': 'dictor',
        'NOISE': 'noise',
    }
    assert Question.labels() == {
        Question.DICTOR: {'EN': 'Is dictor the same for both audios?', 'RU': 'На записях звучит один и тот же голос?'},
        Question.NOISE: {'EN': 'Is there any noise in the audio?', 'RU': 'На записи присутствует шум?'},
    }

    assert question_answers == [
        (Question.DICTOR, [Answer.DICTOR__SAME, Answer.DICTOR__DIFFERENT, Answer.DICTOR__OTHER]),
        (Question.NOISE, [Answer.NOISE__YES, Answer.NOISE__NO]),
    ]



import pandas as pd
import pytest
from typing import Dict

from crowdom import base, objects

from . import lib


def test_display():
    function = base.ClassificationFunction(
        inputs=(base.ObjectMeta(objects.Text, 'my_text'), base.ObjectMeta(objects.Image, 'my_image')),
        cls=lib.ImageClass,
    )
    df = pd.DataFrame(
        [
            {'type': 'Text', 'name': 'my_text'},
            {'type': 'Image', 'name': 'my_image'},
            {'type': 'ImageClass', 'name': 'choice'},
        ]
    )
    assert function.display_arguments(function.get_all_arguments()) == str(df)


class TestValidate:
    def test_correct_classification_no_names(self):
        base.ClassificationFunction(inputs=(objects.Text, objects.Text, objects.Image), cls=lib.ImageClass)

    def test_correct_classification_with_input_names(self):
        base.ClassificationFunction(
            inputs=(
                base.ObjectMeta(objects.Text, 'my_text'),
                base.ObjectMeta(objects.Text, 'other_text'),
                base.ObjectMeta(objects.Image, 'my_image'),
            ),
            cls=lib.ImageClass,
        )

    def test_correct_classification_with_all_names(self):
        base.ClassificationFunction(
            inputs=(
                base.ObjectMeta(objects.Text, 'my_text'),
                base.ObjectMeta(objects.Text, 'other_text'),
                base.ObjectMeta(objects.Image, 'my_image'),
            ),
            cls=base.ClassMeta(lib.ImageClass, 'my_class'),
        )

    def test_correct_sbs_no_names(self):
        base.SbSFunction(inputs=(objects.Text, objects.Text, objects.Image), hints=(objects.Text, objects.Image))

    def test_correct_annotation_no_names(self):
        base.AnnotationFunction(inputs=(objects.Text, objects.Text, objects.Image), outputs=(lib.ImageClass,))

    def test_incorrect_classification_not_object(self):
        with pytest.raises(ValueError) as e:
            base.ClassificationFunction(inputs=(int,), cls=lib.ImageClass)  # noqa
        assert str(e.value) == f'Unexpected type or instance: "{int}"'

    def test_incorrect_classification_regex_empty(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(inputs=(base.ObjectMeta(objects.Text, ''),), cls=lib.ImageClass)

        assert str(e.value) == 'Incorrect name: ""'

    def test_incorrect_classification_regex_long(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(inputs=(base.ObjectMeta(objects.Text, '1' * 31),), cls=lib.ImageClass)

        assert str(e.value) == f'Incorrect name: "{"1" * 31}"'

    def test_incorrect_classification_regex_non_ascii(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(inputs=(base.ObjectMeta(objects.Text, 'abcdeё'),), cls=lib.ImageClass)

        assert str(e.value) == 'Incorrect name: "abcdeё"'

    def test_incorrect_classification_regex_underscore(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(inputs=(base.ObjectMeta(objects.Text, '_my_name'),), cls=lib.ImageClass)

        assert str(e.value) == 'Names starting with "_" are reserved for internal usage'

    def test_incorrect_classification_missing_names_with_class_name(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(
                inputs=(objects.Text, objects.Text, objects.Image), cls=base.ClassMeta(lib.ImageClass, 'my_class')
            )
        df = pd.DataFrame(
            [
                {'type': 'Text', 'name': 'None'},
                {'type': 'Text', 'name': 'None'},
                {'type': 'Image', 'name': 'None'},
                {'type': 'ImageClass', 'name': 'my_class'},
            ]
        )
        assert str(e.value) == f'You should provide names for all types. Current names:\n{str(df)}'

    def test_incorrect_classification_missing_names(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(
                inputs=(
                    base.ObjectMeta(objects.Text, 'my_text'),
                    base.ObjectMeta(objects.Text),
                    base.ObjectMeta(objects.Image, 'my_image'),
                ),
                cls=base.ClassMeta(lib.ImageClass, 'my_class'),
            )
        df = pd.DataFrame(
            [
                {'type': 'Text', 'name': 'my_text'},
                {'type': 'Text', 'name': 'None'},
                {'type': 'Image', 'name': 'my_image'},
                {'type': 'ImageClass', 'name': 'my_class'},
            ]
        )
        assert str(e.value) == f'You should provide names for all types. Current names:\n{str(df)}'

    def test_incorrect_classification_overlapping_input_names(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(
                inputs=(
                    base.ObjectMeta(objects.Text, 'my_text'),
                    base.ObjectMeta(objects.Text, 'my_text'),
                    base.ObjectMeta(objects.Image, 'my_image'),
                ),
                cls=base.ClassMeta(lib.ImageClass, 'my_class'),
            )
        df = pd.DataFrame(
            [
                {'type': 'Text', 'name': 'my_text'},
                {'type': 'Text', 'name': 'my_text'},
                {'type': 'Image', 'name': 'my_image'},
                {'type': 'ImageClass', 'name': 'my_class'},
            ]
        )
        assert str(e.value) == f'Names should be unique. Current names:\n{str(df)}'

    def test_incorrect_classification_overlapping_class_name(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(
                inputs=(
                    base.ObjectMeta(objects.Text, 'my_text'),
                    base.ObjectMeta(objects.Text, 'choice'),
                    base.ObjectMeta(objects.Image, 'my_image'),
                ),
                cls=lib.ImageClass,
            )
        df = pd.DataFrame(
            [
                {'type': 'Text', 'name': 'my_text'},
                {'type': 'Text', 'name': 'choice'},
                {'type': 'Image', 'name': 'my_image'},
                {'type': 'ImageClass', 'name': 'choice'},
            ]
        )
        assert str(e.value) == f'Names should be unique. Current names:\n{str(df)}'

    def test_incorrect_classification_additional_validate_not_class(self):
        with pytest.raises(AssertionError):
            base.ClassificationFunction(inputs=(objects.Image,), cls=objects.Text)  # noqa

    def test_incorrect_classification_additional_validate_not_classmeta(self):
        with pytest.raises(AssertionError):
            meta = base.ObjectMeta(lib.ImageClass, 'my_class')
            base.ClassificationFunction(inputs=(base.ObjectMeta(objects.Text, 'my_text'),), cls=meta)  # noqa

    def test_incorrect_sbs_overlapping_class_name(self):
        with pytest.raises(AssertionError) as e:
            # todo на самом деле это было бы валидно, потому что choice из input'а
            #  трансформировался бы в choice_a, choice_b
            base.SbSFunction(inputs=(base.ObjectMeta(objects.Text, 'my_text'), base.ObjectMeta(objects.Text, 'choice')))
        df = pd.DataFrame(
            [
                {'type': 'Text', 'name': 'my_text'},
                {'type': 'Text', 'name': 'choice'},
                {'type': 'SbSChoice', 'name': 'choice'},
            ]
        )
        assert str(e.value) == f'Names should be unique. Current names:\n{str(df)}'

    def test_incorrect_sbs_additional_validate(self):
        with pytest.raises(NotImplementedError) as e:
            meta = base.SbSChoiceMeta(lib.ImageClass, 'choice_a')
            base.SbSFunction(inputs=(objects.Image,), choice=meta)
        assert str(e.value) == f'Unexpected type or instance: "{meta}"'

    def test_incorrect_annotation_overlapping_class_name(self):
        with pytest.raises(AssertionError) as e:
            base.AnnotationFunction(
                inputs=(base.ObjectMeta(objects.Text, 'source_text'), base.ObjectMeta(objects.Text, 'eval')),
                outputs=(base.ObjectMeta(objects.Text, 'translated_text'),),
            )
        df = pd.DataFrame(
            [
                {'type': 'Text', 'name': 'source_text'},
                {'type': 'Text', 'name': 'eval'},
                {'type': 'Text', 'name': 'translated_text'},
                {'type': 'BinaryEvaluation', 'name': 'eval'},
            ]
        )
        assert str(e.value) == f'Names should be unique. Current names:\n{str(df)}'

    def test_incorrect_annotation_additional_validate(self):
        with pytest.raises(NotImplementedError) as e:
            meta = base.EvaluationMeta(lib.ImageClass, 'choice')
            base.AnnotationFunction(
                inputs=(base.ObjectMeta(objects.Text, 'source_text'),),
                outputs=(base.ObjectMeta(objects.Text, 'translated_text'),),
                evaluation=meta,
            )  # noqa
        assert str(e.value) == f'Unexpected type: "{meta}"'

    def test_incorrect_classification_optional_choice(self):
        with pytest.raises(AssertionError) as e:
            base.ClassificationFunction(
                inputs=(base.ObjectMeta(objects.Text),),
                cls=base.ClassMeta(lib.ImageClass, required=False),
            )
        assert str(e.value) == 'choice is required'


class TestLocalizedString:
    def test_concat(self):
        assert base.LocalizedString({'EN': 'one'}) + base.LocalizedString(
            {'EN': ' two', 'DE': ' zwei'}
        ) == base.LocalizedString({'EN': 'one two', 'DE': 'one zwei'})
        assert base.LocalizedString({'EN': 'one', 'DE': 'einer'}) + base.LocalizedString(
            {'EN': ' two'}
        ) == base.LocalizedString({'EN': 'one two', 'DE': 'einer two'})
        assert base.LocalizedString({'EN': 'one', 'DE': 'einer'}) + '.' == base.LocalizedString(
            {'EN': 'one.', 'DE': 'einer.'}
        )
        assert base.LocalizedString({'DE': 'einer', 'EN': 'one', 'RU': 'один'}) + ', ' + base.LocalizedString(
            {'RU': 'два', 'KK': 'екі', 'EN': 'two'}
        ) + '; ' + base.LocalizedString({'EN': 'three', 'RU': 'три'}) == base.LocalizedString(
            {'EN': 'one, two; three', 'RU': 'один, два; три', 'DE': 'einer, two; three', 'KK': 'one, екі; three'}
        )
        with pytest.raises(AssertionError) as e:
            base.LocalizedString({'DE': 'einer'}) + base.LocalizedString({'EN': ' two'})
        assert str(e.value) == 'no text for EN language'

    def test_task_spec(self):
        dict_task_spec = base.TaskSpec(
            id=None,  # noqa
            function=None,  # noqa
            name={'EN': 'en_name', 'RU': 'ru_name'},
            description={'EN': 'en_desc', 'RU': 'ru_desc'},
            instruction={'EN': 'en_instruction', 'RU': 'ru_instruction'},
        )

        ls_task_spec = base.TaskSpec(
            id=None,  # noqa
            function=None,  # noqa
            name=base.LocalizedString({'EN': 'en_name', 'RU': 'ru_name'}),
            description=base.LocalizedString({'EN': 'en_desc', 'RU': 'ru_desc'}),
            instruction=base.LocalizedString({'EN': 'en_instruction', 'RU': 'ru_instruction'}),
        )

        assert dict_task_spec == ls_task_spec

    def test_class(self):
        class Speech(base.Class):
            YES = 'yes'
            NO = 'no'
            OTHER = 'other'

            @classmethod
            def labels(cls) -> Dict['Speech', Dict[str, str]]:
                return {
                    cls.YES: {'EN': 'yes', 'RU': 'да'},
                    cls.NO: {'EN': 'no', 'RU': 'нет'},
                    cls.OTHER: {'EN': 'other', 'RU': 'другое'},
                }

        assert Speech.get_display_labels() == [
            base.LocalizedString({'EN': 'yes', 'RU': 'да'}),
            base.LocalizedString({'EN': 'no', 'RU': 'нет'}),
            base.LocalizedString({'EN': 'other', 'RU': 'другое'}),
        ]


def test_incorrectly_valued_class():
    with pytest.raises(AssertionError) as e:

        class MOS(base.Class):
            ONE = 1
            TWO = 2

        assert str(e.value) == 'only str-valued Classes are accepted'



# We use Protobuf to serialize objects on lzy whiteboards.
# Standard pickle serialization does not fit our needs, because in general case refactorings in library code may lead
# to incorrect objects deserialization for old, already persisted whiteboards.
#
# For each class which will be stored in whiteboards, we define its serialization wrapper. A possible alternative is
# to provide additional metadata (@message class decorator, attributes data, etc.) to source classes, but we fully
# separate serialization logic from pure business-logic classes, because:
# 1) Currently we have issues with `pure-protobuf` dependency in Arcadia contrib.
# 2) Some attributes types, i.e. dicts, are not Protobuf serializable, and we don't want to modify source classes only
#    because of serialization needs.
#
# To serialize classes from inheritance schemes, we use Protobuf one_of.
#
# We also have another case, when client defines his own classes. Such sets of classes are limited (i.e. subclasses of
# Class) and we typically know data of these classes (.value for Class inheritor). But we don't know to which class
# we need to deserialize this data. So we introduce classes registry, and the client have to associate a permanent
# name with each of his self-defined class, so we can later instantiate this concrete class by its name.

import abc
from typing import Any, Callable, Dict, Generic, Type, TypeVar

from ... import base, objects


T = TypeVar('T', bound=Type[Any], covariant=True)
ClassT = TypeVar('ClassT', bound=Type[base.Class], covariant=True)

# DO NOT CHANGE THIS NAMES!
# They are persisted in lzy whiteboards and must remain the same for correct deserialization.
types_registry: Dict[Type[base.Object], str] = {
    base.SbSChoice: 'SbSChoice',
    base.BinaryEvaluation: 'BinaryEvaluation',
    objects.Audio: 'Audio',
    objects.Image: 'Image',
    objects.Text: 'Text',
    objects.Video: 'Video',
}
types_registry_reversed = {name: type for type, name in types_registry.items()}


def register_type(type: Type[base.Object], name: str):
    global types_registry, types_registry_reversed
    types_registry[type] = name
    types_registry_reversed[name] = type


func_registry = {}
func_registry_reversed = {}


def register_func(func: Callable, name: str):
    global func_registry, func_registry_reversed
    func_registry[func] = name
    func_registry_reversed[name] = func


# Most likely we won't need exact version of function as it was implemented at the time of the launch.
# Its outputs will be persisted in whiteboard anyway. So if we can't find it, it's not a problem.
def func_not_found(name: str) -> Callable:
    def f():
        raise RuntimeError(f'function "{name}" not found')

    return f


def load_func(name: str) -> Callable:
    return func_registry_reversed.get(name, func_not_found(name))


def deserialize_one_of_field(field):
    one_of = getattr(field, field.which_one_of)
    if isinstance(one_of, ProtobufSerializer):
        return one_of.deserialize()
    else:
        return one_of  # primitive type


class ProtobufSerializer(Generic[T]):
    @staticmethod
    @abc.abstractmethod
    def serialize(obj: T) -> 'ProtobufSerializer':
        ...

    @abc.abstractmethod
    def deserialize(self) -> T:
        ...



import pytest
from typing import List

from mock import patch
import pandas as pd
import toloka.client as toloka

from crowdom import base, datasource, objects
from . import lib


def test_parse_rows():
    task_mapping = lib.audio_transcript_ext_mapping

    assert datasource.parse_rows(
        [{'audio_link': 'https://storage.net/1.wav'}, {'audio_link': 'https://storage.net/2.wav'}],
        task_mapping,
        has_solutions=False,
    ) == [
        (objects.Audio(url='https://storage.net/1.wav'),),
        (objects.Audio(url='https://storage.net/2.wav'),),
    ]

    assert datasource.parse_rows(
        [
            {'audio_link': 'https://storage.net/1.wav', 'choice': 'sil'},
            {'audio_link': 'https://storage.net/2.wav', 'output': 'hello', 'choice': 'sp'},
        ],
        task_mapping,
        has_solutions=True,
    ) == [
        ((objects.Audio(url='https://storage.net/1.wav'),), (None, lib.sil)),
        (
            (objects.Audio(url='https://storage.net/2.wav'),),
            (
                objects.Text(text='hello'),
                lib.sp,
            ),
        ),
    ]

    no_solutions_err = "error in file entry #1: it must have the following fields: [('audio_link', <class 'str'>)]"

    with pytest.raises(ValueError) as e:
        datasource.parse_rows([{}], task_mapping, has_solutions=False)
    assert str(e.value) == no_solutions_err

    with pytest.raises(ValueError) as e:
        datasource.parse_rows([{'audio': 42}], task_mapping, has_solutions=False)
    assert str(e.value) == no_solutions_err

    has_solutions_err = (
        "error in file entry #1: it must have the following fields: [('audio_link', <class 'str'>), "
        "('output', <class 'str'>), ('choice', <class 'str'>)]"
    )

    with pytest.raises(ValueError) as e:
        datasource.parse_rows([{}], task_mapping, has_solutions=True)
    assert str(e.value) == has_solutions_err

    with pytest.raises(ValueError) as e:
        datasource.parse_rows(
            [{'audio': 'https://storage.net/1.wav', 'output': 'hello'}], task_mapping, has_solutions=True
        )
    assert str(e.value) == has_solutions_err

    with pytest.raises(ValueError) as e:
        datasource.parse_rows(
            [{'audio': 'https://storage.net/1.wav', 'choice': True}], task_mapping, has_solutions=True
        )
    assert str(e.value) == has_solutions_err


def test_file_format():
    task_mapping = lib.audio_transcript_ext_mapping

    pd.testing.assert_frame_equal(
        datasource.file_format(task_mapping, has_solutions=False), pd.DataFrame([{'name': 'audio_link', 'type': 'str'}])
    )
    pd.testing.assert_frame_equal(
        datasource.file_format(task_mapping, has_solutions=True),
        pd.DataFrame(
            [
                {'name': 'audio_link', 'type': 'str'},
                {'name': 'output', 'type': 'str'},
                {'name': 'choice', 'type': 'str'},
            ]
        ),
    )


@patch('boto3.session.Session', lib.Boto3SessionStub)
def test_substitute_media_output():
    assignments = [
        (
            toloka.Assignment(status=toloka.Assignment.REJECTED),
            [
                ((objects.Text(text='hello'),), (objects.Audio(url='attachment-id_1'), objects.Text(text='ok'))),
                ((objects.Text(text='no thank you'),), (objects.Audio(url='attachment-id_2'), None)),
            ],
        ),
        (
            toloka.Assignment(status=toloka.Assignment.ACCEPTED),
            [
                ((objects.Text(text='nice'),), (objects.Audio(url='attachment-id_3'), None)),
                ((objects.Text(text='bye-bye'),), (objects.Audio(url='attachment-id_4'), None)),
            ],
        ),
        (
            toloka.Assignment(status=toloka.Assignment.SUBMITTED),
            [
                ((objects.Text(text='ok'),), (objects.Audio(url='attachment-id_5'), objects.Text(text='ok'))),
                ((objects.Text(text='where is he'),), (objects.Audio(url='attachment-id_6'), objects.Text(text='ok'))),
            ],
        ),
        (
            toloka.Assignment(status=toloka.Assignment.SUBMITTED),
            [
                ((objects.Text(text='later'),), (objects.Audio(url='attachment-id_7'), objects.Text(text='ok'))),
                ((objects.Text(text='no'),), (objects.Audio(url='attachment-id_8'), objects.Text(text='bad'))),
            ],
        ),
    ]

    class TolokaClientStub(lib.TolokaClientCallRecorderStub):
        def __init__(self, attachments: List[toloka.Attachment]):
            self.id_to_attachment = {attachment.id: attachment for attachment in attachments}
            super(TolokaClientStub, self).__init__()

        def get_attachment(self, attachment_id: str) -> toloka.Attachment:
            super(TolokaClientStub, self).get_attachment(attachment_id)
            return self.id_to_attachment[attachment_id]

    s3 = datasource.S3(endpoint='storage.net', bucket='my-bucket', path='Data/Audio')
    stub = TolokaClientStub(
        attachments=[
            toloka.attachment.AssignmentAttachment(id=id, name=name, media_type='application/octet-stream')
            for id, name in [
                ('attachment-id_1', 'aabb.wav'),
                ('attachment-id_2', '1234.wav'),
                ('attachment-id_3', 'qqss.wav'),
                ('attachment-id_4', '5gg5.wav'),
                ('attachment-id_5', '1001.wav'),
                ('attachment-id_6', 'zz99.wav'),
                ('attachment-id_7', 'rr22.wav'),
                ('attachment-id_8', '9876.wav'),
            ]
        ]
    )
    new_assignments = datasource.substitute_media_output(assignments, s3, stub)  # noqa

    assert new_assignments == [
        (
            toloka.Assignment(status=toloka.Assignment.REJECTED),
            [
                (
                    (objects.Text(text='hello'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/aabb.wav'), objects.Text(text='ok')),
                ),
                (
                    (objects.Text(text='no thank you'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/1234.wav'), None),
                ),
            ],
        ),
        (
            toloka.Assignment(status=toloka.Assignment.ACCEPTED),
            [
                (
                    (objects.Text(text='nice'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/qqss.wav'), None),
                ),
                (
                    (objects.Text(text='bye-bye'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/5gg5.wav'), None),
                ),
            ],
        ),
        (
            toloka.Assignment(status=toloka.Assignment.SUBMITTED),
            [
                (
                    (objects.Text(text='ok'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/1001.wav'), objects.Text(text='ok')),
                ),
                (
                    (objects.Text(text='where is he'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/zz99.wav'), objects.Text(text='ok')),
                ),
            ],
        ),
        (
            toloka.Assignment(status=toloka.Assignment.SUBMITTED),
            [
                (
                    (objects.Text(text='later'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/rr22.wav'), objects.Text(text='ok')),
                ),
                (
                    (objects.Text(text='no'),),
                    (objects.Audio(url='https://storage.net/my-bucket/Data/Audio/9876.wav'), objects.Text(text='bad')),
                ),
            ],
        ),
    ]

    assert s3.client.calls == [
        (
            'put_object',
            {
                'ACL': 'public-read',
                'Bucket': 'my-bucket',
                'ContentType': 'application/octet-stream',
                'Key': 'Data/Audio/1001.wav',
            },
        ),
        (
            'put_object',
            {
                'ACL': 'public-read',
                'Bucket': 'my-bucket',
                'ContentType': 'application/octet-stream',
                'Key': 'Data/Audio/zz99.wav',
            },
        ),
        (
            'put_object',
            {
                'ACL': 'public-read',
                'Bucket': 'my-bucket',
                'ContentType': 'application/octet-stream',
                'Key': 'Data/Audio/rr22.wav',
            },
        ),
        (
            'put_object',
            {
                'ACL': 'public-read',
                'Bucket': 'my-bucket',
                'ContentType': 'application/octet-stream',
                'Key': 'Data/Audio/9876.wav',
            },
        ),
    ]

    assert stub.calls == [
        ('get_attachment', ('attachment-id_1',)),
        ('get_attachment', ('attachment-id_2',)),
        ('get_attachment', ('attachment-id_3',)),
        ('get_attachment', ('attachment-id_4',)),
        ('get_attachment', ('attachment-id_5',)),
        ('get_attachment', ('attachment-id_6',)),
        ('get_attachment', ('attachment-id_7',)),
        ('get_attachment', ('attachment-id_8',)),
        ('download_attachment', ('attachment-id_5',)),
        ('download_attachment', ('attachment-id_6',)),
        ('download_attachment', ('attachment-id_7',)),
        ('download_attachment', ('attachment-id_8',)),
    ]


def test_has_media_output():
    assert (
        datasource.has_media_output(base.AnnotationFunction(inputs=(objects.Audio,), outputs=(objects.Text,))) is False
    )
    assert (
        datasource.has_media_output(
            base.AnnotationFunction(inputs=(objects.Text,), outputs=(objects.Text, objects.Image))
        )
        is True
    )


def test_assert_tasks_are_unique():
    with pytest.raises(AssertionError):
        datasource.assert_tasks_are_unique(
            [
                ((objects.Audio(url='https://storage.net/1.wav'),), (lib.sil,)),
                ((objects.Audio(url='https://storage.net/2.wav'),), (lib.sp,)),
                ((objects.Audio(url='https://storage.net/1.wav'),), (lib.sil,)),
            ]
        )


def test_assert_tasks_are_unique_with_metadata():
    datasource.assert_tasks_are_unique(
        [
            (objects.Audio(url='https://storage.net/1.wav'), base.Metadata('metadata-1')),
            (objects.Audio(url='https://storage.net/2.wav'), base.Metadata('metadata-2')),
            (objects.Audio(url='https://storage.net/1.wav'), base.Metadata('metadata-3')),
        ]
    )



from dataclasses import dataclass

from pure_protobuf.dataclasses_ import field, message, one_of, part
from pure_protobuf.oneof import OneOf_

from ... import pricing
from .common import ProtobufSerializer, deserialize_one_of_field


@message
@dataclass
class StaticPricingStrategy(ProtobufSerializer[pricing.StaticPricingStrategy]):
    @staticmethod
    def serialize(obj: pricing.StaticPricingStrategy) -> 'StaticPricingStrategy':
        return StaticPricingStrategy()

    def deserialize(self) -> pricing.StaticPricingStrategy:
        return pricing.StaticPricingStrategy()


@message
@dataclass
class DynamicPricingStrategy(ProtobufSerializer[pricing.DynamicPricingStrategy]):
    max_ratio: float = field(1)

    @staticmethod
    def serialize(obj: pricing.DynamicPricingStrategy) -> 'DynamicPricingStrategy':
        return DynamicPricingStrategy(obj.max_ratio)

    def deserialize(self) -> pricing.DynamicPricingStrategy:
        return pricing.DynamicPricingStrategy(self.max_ratio)


@message
@dataclass
class PricingStrategy(ProtobufSerializer[pricing.PricingStrategy]):
    strategy: OneOf_ = one_of(
        static=part(StaticPricingStrategy, 1),
        dynamic=part(DynamicPricingStrategy, 2),
    )

    @staticmethod
    def serialize(obj: pricing.PricingStrategy) -> 'PricingStrategy':
        strategy = PricingStrategy()
        if isinstance(obj, pricing.StaticPricingStrategy):
            strategy.strategy.static = StaticPricingStrategy.serialize(obj)
        elif isinstance(obj, pricing.DynamicPricingStrategy):
            strategy.strategy.dynamic = DynamicPricingStrategy.serialize(obj)
        else:
            raise ValueError(f'unexpected pricing strategy type: {type(obj)}')
        return strategy

    def deserialize(self) -> pricing.PricingStrategy:
        return deserialize_one_of_field(self.strategy)



from collections import defaultdict

from pytest import approx
import toloka.client as toloka

from crowdom import classification, objects, worker
from . import lib

bob, alice, john, mary = (
    worker.Human(toloka.Assignment(user_id=user_id)) for user_id in ('bob', 'alice', 'john', 'mary')
)


def test_label_prediction_empty_labels():
    # it should not raise any Exceptions
    for aggregation_algorithm in list(classification.AggregationAlgorithm):
        assert classification.predict_labels_probas_for_class(
            labels=[[], []], aggregation_algorithm=aggregation_algorithm, worker_weights_and_classes=({}, [])
        ) == [None, None]


def test_label_prediction():
    cat, dog = lib.cat, lib.dog

    expected = [None, {cat: 1 / 3, dog: 2 / 3}, {cat: 1, dog: 0}]
    actual = classification.predict_labels_probas(
        labels=[
            [],
            [
                (dog, bob),
                (dog, alice),
                (cat, john),
            ],
            [
                (cat, john),
                (cat, bob),
                (cat, alice),
            ],
        ],
        aggregation_algorithm=classification.AggregationAlgorithm.MAJORITY_VOTE,
        task_mapping=lib.image_classification_mapping,
    )
    assert actual == expected


def test_label_prediction_with_same_weights():
    cat, dog, crow = lib.ImageClass.possible_instances()

    expected = [
        None,
        {cat: approx(2 / 7), dog: approx(4 / 7), crow: approx(1 / 7)},
        {cat: approx(8 / 10), dog: approx(1 / 10), crow: approx(1 / 10)},
    ]
    actual = classification.predict_labels_probas(
        labels=[
            [],
            [
                (dog, bob),
                (dog, alice),
                (cat, john),
            ],
            [
                (cat, john),
                (cat, bob),
                (cat, alice),
            ],
        ],
        aggregation_algorithm=classification.AggregationAlgorithm.MAX_LIKELIHOOD,
        task_mapping=lib.image_classification_mapping,
        worker_weights=defaultdict(lambda: 0.5),
    )
    assert actual == expected


def test_label_prediction_with_different_weights():
    cat, dog, crow = lib.ImageClass.possible_instances()

    expected = [
        {cat: approx(486 / 521), dog: approx(8 / 521), crow: approx(27 / 521)},
        {cat: approx(648 / 681), dog: approx(6 / 681), crow: approx(27 / 681)},
    ]
    actual = classification.predict_labels_probas(
        labels=[
            [
                (dog, bob),
                (dog, john),
                (cat, alice),
            ],
            [
                (dog, bob),
                (cat, john),
                (cat, alice),
            ],
        ],
        aggregation_algorithm=classification.AggregationAlgorithm.MAX_LIKELIHOOD,
        task_mapping=lib.image_classification_mapping,
        worker_weights={bob.id: 0.1, alice.id: 0.9, john.id: 0.4},
    )
    assert actual == expected


def test_combined_class_label_prediction():
    dictor_same, dictor_different, dictor_other, noise_yes, noise_no = lib.Answer.possible_instances()

    expected = [
        {dictor_different: approx(3 / 19), dictor_other: approx(4 / 19), dictor_same: approx(12 / 19)},
        {noise_no: approx(0.4), noise_yes: approx(0.6)},
    ]
    actual = classification.predict_labels_probas(
        labels=[
            [
                (dictor_same, bob),
                (dictor_other, john),
                (dictor_same, alice),
            ],
            [
                (noise_yes, bob),
                (noise_no, john),
                (noise_yes, alice),
            ],
        ],
        aggregation_algorithm=classification.AggregationAlgorithm.MAX_LIKELIHOOD,
        task_mapping=lib.question_answer_mapping,
        worker_weights={bob.id: 0.1, alice.id: 0.9, john.id: 0.4},
    )
    assert actual == expected


def test_most_probable_label():
    cat, dog, crow = lib.cat, lib.dog, lib.crow
    assert classification.get_most_probable_label({cat: 0.1, dog: 0.5, crow: 0.4}) == (dog, 0.5)
    assert classification.get_most_probable_label(None) is None


# TODO: test Class-valued rows in df with Dawid Skene crowdkit implementation
def test_labels_probas_from_assignments():
    images = [objects.Image(url=f'https://storage.net/{i + 1}.jpg') for i in range(5)]
    control_image = objects.Image(url='https://storage.net/42.jpg')
    cat, dog, crow = lib.cat, lib.dog, lib.crow
    image_class_controls = [(control_image, cat)]

    bob_a, john_a, mary_a, alice_a = (
        lib.create_classification_assignment(image_class_pairs, image_class_controls, user_id=user_id)[0]
        for image_class_pairs, user_id in (
            ([(images[4], dog), (images[0], cat), (control_image, dog), (images[2], cat), (images[1], cat)], 'bob'),
            ([(images[2], dog), (images[1], cat), (images[4], dog), (images[3], crow)], 'john'),
            ([(images[0], cat), (images[2], crow), (images[4], crow), (images[1], cat)], 'mary'),
            ([(control_image, cat), (images[3], dog), (images[1], cat), (images[0], dog)], 'alice'),
        )
    )

    bob_, john_, mary_, alice_ = (worker.Human(assignment) for assignment in (bob_a, john_a, mary_a, alice_a))

    expected = [
        ({dog: 1 / 3, cat: 2 / 3, crow: 0.0}, [(cat, bob_), (cat, mary_), (dog, alice_)]),
        ({dog: 0.0, cat: 4 / 4, crow: 0.0}, [(cat, bob_), (cat, john_), (cat, mary_), (cat, alice_)]),
        ({dog: 1 / 3, cat: 1 / 3, crow: 1 / 3}, [(cat, bob_), (dog, john_), (crow, mary_)]),
        ({dog: 1 / 2, cat: 0.0, crow: 1 / 2}, [(crow, john_), (dog, alice_)]),
        ({dog: 2 / 3, cat: 0.0, crow: 1 / 3}, [(dog, bob_), (dog, john_), (crow, mary_)]),
    ]

    assert (
        classification.collect_labels_probas_from_assignments(
            assignments=[bob_a, john_a, mary_a, alice_a],
            task_mapping=lib.image_classification_mapping,
            pool_input_objects=[(image,) for image in images],
            aggregation_algorithm=classification.AggregationAlgorithm.MAJORITY_VOTE,
        )
        == expected
    )



from dataclasses import dataclass
from typing import Optional

from pure_protobuf.dataclasses_ import field, message

from ... import evaluation
from .common import ProtobufSerializer


@message
@dataclass
class AssignmentCheckSample(ProtobufSerializer[evaluation.AssignmentCheckSample]):
    max_tasks_to_check: Optional[int] = field(1, default=None)
    assignment_accuracy_finalization_threshold: Optional[float] = field(2, default=None)

    @staticmethod
    def serialize(obj: evaluation.AssignmentCheckSample) -> 'AssignmentCheckSample':
        return AssignmentCheckSample(
            max_tasks_to_check=obj.max_tasks_to_check,
            assignment_accuracy_finalization_threshold=obj.assignment_accuracy_finalization_threshold,
        )

    def deserialize(self) -> evaluation.AssignmentCheckSample:
        return evaluation.AssignmentCheckSample(
            max_tasks_to_check=self.max_tasks_to_check,
            assignment_accuracy_finalization_threshold=self.assignment_accuracy_finalization_threshold,
        )
